import logging
import multiprocessing as mp
import os
import pickle
from abc import ABC, abstractmethod
from time import sleep
from typing import Dict
import shutil
import numpy as np

from autoconf import conf
from autofit import exc
from autofit.mapper import model_mapper as mm
from autofit.non_linear.initializer import Initializer
from autofit.non_linear.log import logger
from autofit.non_linear.paths import Paths, convert_paths
from autofit.non_linear.timer import Timer
from autofit.text import formatter
from autofit.text import model_text
from autofit.text import samples_text


class NonLinearSearch(ABC):
    @convert_paths
    def __init__(
            self,
            paths=None,
            prior_passer=None,
            initializer=None,
            iterations_per_update=None,
            number_of_cores=1,
    ):
        """Abstract base class for non-linear searches.

        This class sets up the file structure for the non-linear search, which are standardized across all non-linear
        searches.

        Parameters
        ------------
        paths : af.Paths
            Manages all paths, e.g. where the search outputs are stored, the samples, backups, etc.
        prior_passer : af.PriorPasser
            Controls how priors are passed from the results of this non-linear search to a subsequent non-linear search.
        initializer : non_linear.initializer.Initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        """

        if paths.non_linear_name is "":
            paths.non_linear_name = self._config("tag", "name")

        if paths.non_linear_tag is "":
            paths.non_linear_tag_function = lambda: self.tag

        self.paths = paths
        if prior_passer is None:
            self.prior_passer = PriorPasser.from_config(config=self._config)
        else:
            self.prior_passer = prior_passer

        self.timer = Timer(paths=paths)

        self.skip_completed = conf.instance.general.get(
            "output", "skip_completed", bool
        )
        self.force_pickle_overwrite = conf.instance.general.get(
            "output", "force_pickle_overwrite", bool
        )

        self.log_file = conf.instance.general.get("output", "log_file", str).replace(
            " ", ""
        )

        if initializer is None:
            self.initializer = Initializer.from_config(config=self._config)
        else:
            self.initializer = initializer

        self.iterations_per_update = (
            self._config("updates", "iterations_per_update", int)
            if iterations_per_update is None
            else iterations_per_update
        )

        if conf.instance.general.get("hpc", "hpc_mode", bool):
            self.iterations_per_update = conf.instance.general.get("hpc", "iterations_per_update", float)

        self.log_every_update = self._config("updates", "log_every_update", int)
        self.backup_every_update = self._config("updates", "backup_every_update", int)
        self.visualize_every_update = self._config(
            "updates", "visualize_every_update", int
        )
        self.model_results_every_update = self._config(
            "updates", "model_results_every_update", int
        )

        self.iterations = 0
        self.should_log = IntervalCounter(self.log_every_update)
        self.should_backup = IntervalCounter(self.backup_every_update)
        self.should_visualize = IntervalCounter(self.visualize_every_update)
        self.should_output_model_results = IntervalCounter(
            self.model_results_every_update
        )

        self.silence = self._config("printing", "silence", bool)

        if conf.instance.general.get("hpc", "hpc_mode", bool):
            self.silence = True

        self.number_of_cores = number_of_cores

        self._in_phase = False

    class Fitness:
        def __init__(self, paths, model, analysis, samples_from_model, pool_ids=None):

            self.paths = paths
            self.max_log_likelihood = -np.inf
            self.analysis = analysis

            self.model = model
            self.samples_from_model = samples_from_model

            self.pool_ids = pool_ids

        def fit_instance(self, instance):

            log_likelihood = self.analysis.log_likelihood_function(instance=instance)

            if self.analysis.log_likelihood_cap is not None:
                if log_likelihood > self.analysis.log_likelihood_cap:
                    log_likelihood = self.analysis.log_likelihood_cap

            if log_likelihood > self.max_log_likelihood:

                if self.pool_ids is not None:
                    if mp.current_process().pid != min(self.pool_ids):
                        return log_likelihood

                self.max_log_likelihood = log_likelihood

            return log_likelihood

        def log_likelihood_from_parameters(self, parameters):
            instance = self.model.instance_from_vector(vector=parameters)
            log_likelihood = self.fit_instance(instance)
            return log_likelihood

        def log_posterior_from_parameters(self, parameters):
            log_likelihood = self.log_likelihood_from_parameters(parameters=parameters)
            log_priors = self.model.log_priors_from_vector(vector=parameters)
            return log_likelihood + sum(log_priors)

        def figure_of_merit_from_parameters(self, parameters):
            """The figure of merit is the value that the non-linear search uses to sample parameter space. This varies
            between different non-linear search algorithms, for example:

                - The *Optimizer* *PySwarms* uses the chi-squared value, which is the -2.0*log_posterior.
                - The *MCMC* algorithm *Emcee* uses the log posterior.
                - Nested samplers such as *Dynesty* use the log likelihood.
            """
            raise NotImplementedError()

        @staticmethod
        def prior(cube, model):

            # NEVER EVER REFACTOR THIS LINE! Haha.

            phys_cube = model.vector_from_unit_vector(unit_vector=cube)

            for i in range(len(phys_cube)):
                cube[i] = phys_cube[i]

            return cube

        @staticmethod
        def fitness(cube, model, fitness_function):
            return fitness_function(instance=model.instance_from_vector(cube))

        @property
        def samples(self):
            return self.samples_from_model(model=self.model)

        @property
        def resample_figure_of_merit(self):
            """If a sample raises a FitException, this value is returned to signify that the point requires resampling or
             should be given a likelihood so low that it is discard."""
            return -np.inf

    def fit(self, model, analysis: "Analysis", info=None, pickle_files=None) -> "Result":
        """ Fit a model, M with some function f that takes instances of the
        class represented by model M and gives a score for their fitness.

        A model which represents possible instances with some dimensionality is fit.

        The analysis provides two functions. One visualises an instance of a model and the
        other scores an instance based on how well it fits some data. The search
        produces instances of the model by picking points in an N dimensional space.

        Parameters
        ----------
        analysis : af.Analysis
            An object that encapsulates the data and a log likelihood function.
        model : ModelMapper
            An object that represents possible instances of some model with a
            given dimensionality which is the number of free dimensions of the
            model.
        info : dict
            Optional dictionary containing information about the fit that can be loaded by the aggregator.
        pickle_files : [str]
            Optional list of strings specifying the path and filename of .pickle files, that are copied to each
            model-fits pickles folder so they are accessible via the Aggregator.

        Returns
        -------
        An object encapsulating how well the model fit the data, the best fit instance
        and an updated model with free parameters updated to represent beliefs
        produced by this fit.
        """

        self.paths.restore()
        self.setup_log_file()

        if (not os.path.exists(self.paths.has_completed_path) or not self.skip_completed) or self.force_pickle_overwrite:

            self.save_model_info(model=model)
            self.save_parameter_names_file(model=model)
            self.save_metadata()
            self.save_info(info=info)
            self.save_search()
            self.save_model(model=model)
            self.move_pickle_files(pickle_files=pickle_files)

        if not os.path.exists(self.paths.has_completed_path) or not self.skip_completed:

            # TODO : Better way to handle?
            self.timer.paths = self.paths
            self.timer.start()

            self._fit(model=model, analysis=analysis)
            open(self.paths.has_completed_path, "w+").close()

            samples = self.perform_update(
                model=model, analysis=analysis, during_analysis=False
            )

        else:

            logger.info(f"{self.paths.name} already completed, skipping non-linear search.")
            samples = self.samples_via_csv_json_from_model(model=model)
            self.save_samples(samples=samples)

        self.paths.backup_zip_remove()

        return Result(samples=samples, previous_model=model, search=self)

    @abstractmethod
    def _fit(self, model, analysis):
        pass

    @property
    def tag(self):
        """Tag the output folder of the non-linear search, based on the non linear search settings"""
        raise NotImplementedError

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        name = "{}/{}".format(self.paths.name, extension)

        if remove_phase_tag:
            tag = ""
        else:
            tag = self.paths.tag

        new_instance = self.__class__(
            paths=Paths(
                name=name,
                tag=tag,
                folders=self.paths.folders,
                path_prefix=self.paths.path_prefix,
                non_linear_name=self.paths.non_linear_name,
                remove_files=self.paths.remove_files,
            )
        )

        return new_instance

    @property
    def config_type(self):
        raise NotImplementedError()

    def _config(self, section, attribute_name, attribute_type=str):
        """
        Get a config field from this search's section in non_linear.ini by a key and value type.

        Parameters
        ----------
        attribute_name: str
            The analysis_path of the field
        attribute_type: type
            The type of the value

        Returns
        -------
        attribute
            An attribute for the key with the specified type.
        """
        return self.config_type.config_for(self.__class__.__name__).get(
            section, attribute_name, attribute_type
        )

    def perform_update(self, model, analysis, during_analysis):
        """Perform an update of the non-linear search results, which occurs every *iterations_per_update* of the
        non-linear search. The update performs the following tasks:

        1) Visualize the maximum log likelihood model.
        2) Backup the samples.
        3) Output the model results to the model.reults file.

        These task are performed every n updates, set by the relevent *task_every_update* variable, for example
        *visualize_every_update* and *backup_every_update*.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the non-linear search maximizes.
        during_analysis : bool
            If the update is during a non-linear search, in which case tasks are only performed after a certain number
             of updates and only a subset of visualization may be performed.
        """

        self.iterations += self.iterations_per_update
        logger.info(f"{self.iterations} Iterations: Performing update (Visualization, outputting samples, etc.).")

        if self.should_backup() or not during_analysis:
            self.paths.backup()

        self.timer.update()

        samples = self.samples_via_sampler_from_model(model=model)
        samples.write_table(filename=f"{self.paths.sym_path}/samples.csv")
        samples.info_to_json(filename=f"{self.paths.sym_path}/info.json")

        self.save_samples(samples=samples)

        try:
            instance = samples.max_log_likelihood_instance
        except exc.FitException:
            return samples

        if self.should_visualize() or not during_analysis:
            analysis.visualize(instance=instance, during_analysis=during_analysis)

        if self.should_output_model_results() or not during_analysis:

            samples_text.results_to_file(
                samples=samples,
                filename=self.paths.file_results,
                during_analysis=during_analysis,
            )

            samples_text.search_summary_to_file(samples=samples, filename=self.paths.file_search_summary)

        return samples

    def setup_log_file(self):

        if conf.instance.general.get("output", "log_to_file", bool):

            if len(self.log_file) == 0:
                raise ValueError("In general.ini log_to_file is True, but log_file is an empty string. "
                                 "Either give log_file a name or set log_to_file to False.")

            log_path = "{}/{}".format(self.paths.output_path, self.log_file)
            logger.handlers = [logging.FileHandler(log_path)]
            logger.propagate = False

    @property
    def samples_cls(self):
        raise NotImplementedError()

    def save_model_info(self, model):
        """Save the model.info file, which summarizes every parameter and prior."""
        with open(self.paths.file_model_info, "w+") as f:
            f.write(model.info)

    def save_parameter_names_file(self, model):
        """Create the param_names file listing every parameter's label and Latex tag, which is used for *GetDist*
        visualization.

        The parameter labels are determined using the label.ini and label_format.ini config files."""

        paramnames_names = model.parameter_names
        paramnames_labels = model_text.parameter_labels_from_model(model=model)

        parameter_name_and_label = []

        for i in range(model.prior_count):
            line = formatter.label_and_label_string(
                label0=paramnames_names[i], label1=paramnames_labels[i], whitespace=70
            )
            parameter_name_and_label += [f"{line}\n"]

        formatter.output_list_of_strings_to_file(
            file=self.paths.file_param_names, list_of_strings=parameter_name_and_label
        )

    def save_info(self, info):
        """
        Save the dataset associated with the phase
        """
        with open("{}/info.pickle".format(self.paths.pickle_path), "wb") as f:
            pickle.dump(info, f)

    def save_search(self):
        """
        Save the seawrch associated with the phase as a pickle
        """
        with open(self.paths.make_search_pickle_path(), "w+b") as f:
            f.write(pickle.dumps(self))

    def save_model(self, model):
        """
        Save the model associated with the phase as a pickle
        """
        with open(self.paths.make_model_pickle_path(), "w+b") as f:
            f.write(pickle.dumps(model))

    def save_samples(self, samples):
        """
        Save the final-result samples associated with the phase as a pickle
        """
        with open(self.paths.make_samples_pickle_path(), "w+b") as f:
            f.write(pickle.dumps(samples))

    def save_metadata(self):
        """
        Save metadata associated with the phase, such as the name of the pipeline, the
        name of the phase and the name of the dataset being fit
        """
        with open("{}/metadata".format(self.paths.make_path()), "a") as f:
            f.write(self.make_metadata_text())

    def move_pickle_files(self, pickle_files):
        """
        Move extra files a user has input the full path + filename of from the location specified to the
        pickles folder of the Aggregator, so that they can be accessed via the aggregator.
        """
        if pickle_files is not None:
            [shutil.copy(file, self.paths.pickle_path) for file in pickle_files]

    @property
    def _default_metadata(self) -> Dict[str, str]:
        """
        A dictionary of metadata describing this phase, including the pipeline
        that it's embedded in.
        """
        return {
            "name": self.paths.name,
            "tag": self.paths.tag,
            "non_linear_search": type(self).__name__.lower(),
        }

    def make_metadata_text(self):
        return "\n".join(
            f"{key}={value or ''}" for key, value in {**self._default_metadata}.items()
        )

    def samples_via_sampler_from_model(self, model):
        raise NotImplementedError()

    def samples_via_csv_json_from_model(self, model):
        raise NotImplementedError()

    def make_pool(self):
        """Make the pool instance used to parallelize a non-linear search alongside a set of unique ids for every
        process in the pool. If the specified number of cores is 1, a pool instance is not made and None is returned.

        The pool cannot be set as an attribute of the class itself because this prevents pickling, thus it is generated
        via this function before calling the non-linear search.

        The pool instance is also set up with a list of unique pool ids, which are used during model-fitting to
        identify a 'master core' (the one whose id value is lowest) which handles model result output, visualization,
        etc."""

        if self.number_of_cores == 1:

            return None, None

        else:

            manager = mp.Manager()
            idQueue = manager.Queue()

            [idQueue.put(i) for i in range(self.number_of_cores)]

            pool = mp.Pool(
                processes=self.number_of_cores, initializer=init, initargs=(idQueue,)
            )
            ids = pool.map(f, range(self.number_of_cores))

            return pool, [id[1] for id in ids]

    def __eq__(self, other):
        return isinstance(other, NonLinearSearch) and self.__dict__ == other.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.paths.restore()


class Analysis:
    def __init__(self, log_likelihood_cap=None):
        self.log_likelihood_cap = log_likelihood_cap

    def log_likelihood_function(self, instance):
        raise NotImplementedError()

    def visualize(self, instance, during_analysis):
        pass


class Result:
    """
    @DynamicAttrs
    """

    def __init__(self, samples, previous_model, search=None):
        """
        The result of an optimization.

        Parameters
        ----------
        previous_model
            The model mapper from the stage that produced this result
        prior_passer : af.PriorPasser
            Controls how priors are passed from the results of this non-linear search to a subsequent non-linear search.
        """

        self.samples = samples
        self.previous_model = previous_model
        self.search = search

        self.__model = None

        self._instance = (
            samples.max_log_likelihood_instance if samples is not None else None
        )

    @property
    def log_likelihood(self):
        return max(self.samples.log_likelihoods)

    @property
    def instance(self):
        return self._instance

    @property
    def max_log_likelihood_instance(self):
        return self._instance

    @property
    def model(self):
        if self.__model is None:
            tuples = self.samples.gaussian_priors_at_sigma(
                sigma=self.search.prior_passer.sigma
            )
            self.__model = self.previous_model.mapper_from_gaussian_tuples(
                tuples,
                use_errors=self.search.prior_passer.use_errors,
                use_widths=self.search.prior_passer.use_widths
            )
        return self.__model

    @model.setter
    def model(self, model):
        self.__model = model

    def __str__(self):
        return "Analysis Result:\n{}".format(
            "\n".join(
                ["{}: {}".format(key, value) for key, value in self.__dict__.items()]
            )
        )

    def model_absolute(self, a: float) -> mm.ModelMapper:
        """
        Parameters
        ----------
        a
            The absolute width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this phase and creating priors with the defined absolute
        width.
        """
        return self.previous_model.mapper_from_gaussian_tuples(
            self.samples.gaussian_priors_at_sigma(sigma=self.search.prior_passer.sigma), a=a
        )

    def model_relative(self, r: float) -> mm.ModelMapper:
        """
        Parameters
        ----------
        r
            The relative width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this phase and creating priors with the defined relative
        width.
        """
        return self.previous_model.mapper_from_gaussian_tuples(
            self.samples.gaussian_priors_at_sigma(sigma=self.search.prior_passer.sigma), r=r
        )


class IntervalCounter:
    def __init__(self, interval):
        self.count = 0
        self.interval = interval

    def __call__(self):
        if self.interval == -1:
            return False
        self.count += 1
        return self.count % self.interval == 0


class PriorPasser:

    def __init__(self, sigma, use_errors, use_widths):
        """Class to package the API for prior passing.

        This class contains the parameters that controls how priors are passed from the results of one non-linear
        search to the next.

        Using the Phase API, we can pass priors from the result of one phase to another follows:

            model_component.parameter = phase1_result.model.model_component.parameter

        By invoking the 'model' attribute, the prior is passed following 3 rules:

            1) The new parameter uses a GaussianPrior. A GaussianPrior is ideal, as the 1D pdf results we compute at
               the end of a phase are easily summarized as a Gaussian.

            2) The mean of the GaussianPrior is the median PDF value of the parameter estimated in phase 1.

              This ensures that the initial sampling of the new phase's non-linear starts by searching the region of
              non-linear parameter space that correspond to highest log likelihood solutions in the previous phase.
              Thus, we're setting our priors to look in the 'correct' regions of parameter space.

            3) The sigma of the Gaussian will use the maximum of two values:

                    (i) the 1D error of the parameter computed at an input sigma value (default sigma=3.0).
                    (ii) The value specified for the profile in the 'config/json_priors/*.json' config
                         file's 'width_modifer' field (check these files out now).

               The idea here is simple. We want a value of sigma that gives a GaussianPrior wide enough to search a
               broad region of parameter space, so that the model can change if a better solution is nearby. However,
               we want it to be narrow enough that we don't search too much of parameter space, as this will be slow or
               risk leading us into an incorrect solution! A natural choice is the errors of the parameter from the
               previous phase.

               Unfortunately, this doesn't always work. Modeling can be prone to an effect called 'over-fitting' where
               we underestimate the parameter errors. This is especially true when we take the shortcuts in early
               phases - fast non-linear search settings, simplified models, etc.

               Therefore, the 'width_modifier' in the json config files are our fallback. If the error on a parameter
               is suspiciously small, we instead use the value specified in the widths file. These values are chosen
               based on our experience as being a good balance broadly sampling parameter space but not being so narrow
               important solutions are missed.

        There are two ways a value is specified using the priors/width file:

            1) Absolute: In this case, the error assumed on the parameter is the value given in the config file. For
               example, if for the width on the parameter of a model component the width modifier reads "Absolute" with
               a value 0.05. This means if the error on the parameter was less than 0.05 in the previous phase, the
               sigma of its GaussianPrior in this phase will be 0.05.

            2) Relative: In this case, the error assumed on the parameter is the % of the value of the estimate value
               given in the config file. For example, if the parameter estimated in the previous phase was 2.0, and the
               relative error in the config file reads "Relative" with a value 0.5, then the sigma of the GaussianPrior
               will be 50% of this value, i.e. sigma = 0.5 * 2.0 = 1.0.

        The PriorPasser allows us to customize at what sigma the error values the model results are computed at to
        compute the passed sigma values and customizes whether the widths in the config file, these computed errors,
        or both, are used to set the sigma values of the passed priors.

        The default values of the PriorPasser are found in the config file of every non-linear search, in the
        [prior_passer] section. All non-linear searches by default use a sigma value of 3.0, use_width=True and
        use_errors=True. We anticipate you should not need to change these values to get lens modeling to work
        proficiently!

        Example:

        Lets say in phase 1 we fit a model, and we estimate that a parameter is equal to 4.0 +- 2.0, where the error
        value of 2.0 was computed at 3.0 sigma confidence. To pass this as a prior to phase 2, we would write:

            model_component.parameter = phase1.result.model.model_component.parameter

        The prior on the parameter in phase 2 would thus be a GaussianPrior, with mean=4.0 and
        sigma=2.0. If we had used a sigma value of 1.0 to compute the error, which reduced the estimate from 4.0 +- 2.0
        to 4.0 +- 0.5, the sigma of the Gaussian prior would instead be 0.5.

        If the error on the parameter in phase 1 had been really small, lets say, 0.01, we would instead use the value
        of the parameter width in the json_priors config file to set sigma instead. Lets imagine the prior config file
        specifies that we use an "Absolute" value of 0.8 to link this prior. Then, the GaussianPrior in phase 2 would
        have a mean=4.0 and sigma=0.8.

        If the prior config file had specified that we use an relative value of 0.8, the GaussianPrior in phase 2 would
        have a mean=4.0 and sigma=3.2.
        """

        self.sigma = sigma
        self.use_errors = use_errors
        self.use_widths = use_widths

    @classmethod
    def from_config(cls, config):
        """Load the PriorPasser from a non_linear config file."""
        sigma = config("prior_passer", "sigma", float)
        use_errors = config("prior_passer", "use_errors", bool)
        use_widths = config("prior_passer", "use_widths", bool)
        return PriorPasser(sigma=sigma, use_errors=use_errors, use_widths=use_widths)


def init(queue):
    global idx
    idx = queue.get()


def f(x):
    global idx
    process = mp.current_process()
    sleep(1)
    return (idx, process.pid, x * x)



from howtofit.chapter_2_results.src.dataset import dataset as ds
from howtofit.chapter_2_results.src.fit import fit as f

from functools import partial

# This module contains convenience methods for computing standard objects from an *Aggregator* as generators.

# You should read this class in detail once you come to th end of part of tutorial 8, as the functions are written in a
# specific way such that they can be used as generators will become clear once the tutorial is completed.


def masked_dataset_generator_from_aggregator(aggregator, settings_masked_dataset=None):
    """Compute a generator of *MaskedDataset* objects from an input aggregator, which generates a list of the 
    *MaskedDataset* objects for every set of results loaded in the aggregator.

    This is performed by mapping the *masked_dataset_from_agg_obj* with the aggregator, which sets up each masked
    _Dataset_ using only generators ensuring that manipulating the _MaskedDataset_ objects of large sets of results is
    done in a memory efficient way.

    Parameters
    ----------
    aggregator : af.Aggregator
        A PyAutoFit aggregator object containing the results of PyAutoGalaxy model-fits."""

    func = partial(
        masked_dataset_from_agg_obj, settings_masked_dataset=settings_masked_dataset
    )

    return aggregator.map(func=func)


def masked_dataset_from_agg_obj(agg_obj, settings_masked_dataset=None):
    """Compute a *MaskedDataset* object from an aggregator's *PhaseOutput* class, which we call an 'agg_obj' to describe 
     that it acts as the aggregator object for one result in the *Aggregator*. This uses the aggregator's generator 
     outputs such that the function can use the *Aggregator*'s map function to to create a *MaskedDataset* generator.

     The *MaskedDataset* is created following the same method as the *Phase* classes, including using the
     *meta_dataset* instance output by the phase to load inputs of the *MaskedDataset* (e.g. data_trim_left). 

    Parameters
    ----------
    agg_obj : af.PhaseOutput
        A PyAutoFit aggregator's PhaseOutput object containing the generators of the results of model-fits.
    """

    dataset = agg_obj.dataset
    mask = agg_obj.mask
    settings = agg_obj.settings

    if settings_masked_dataset is None:
        settings_masked_dataset = settings.settings_masked_dataset

    return ds.MaskedDataset(
        dataset=dataset, mask=mask, settings=settings_masked_dataset
    )


def model_data_generator_from_aggregator(aggregator, settings_masked_dataset=None):
    """Compute a generator of the model data arrays of the 1D profile models from an input aggregator, which
    generates a list of 1D ndarrays every set of results loaded in the aggregator.

    This is performed by mapping the *model_data_from_agg_obj* with the aggregator, which sets up each model data array
    using  only generators ensuring that manipulating the profiles of large sets of results is done in a memory
    efficient way.

    Parameters
    ----------
    aggregator : af.Aggregator
        A PyAutoFit aggregator object containing the results of PyAutoGalaxy model-fits."""

    func = partial(
        model_data_from_agg_obj, settings_masked_dataset=settings_masked_dataset
    )

    return aggregator.map(func=func)


def model_data_from_agg_obj(agg_obj, settings_masked_dataset=None):
    """Compute model data as a 1D ndarray from an aggregator's *PhaseOutput* class, which we call an 'agg_obj' to 
    describe that it acts as the aggregator object for one result in the *Aggregator*. This uses the aggregator's 
    generator outputs such that the function can use the *Aggregator*'s map function to to create a model-data generator.

     The model-data is created following the same method as the *Analysis* classes using an instance of the maximum 
     log likelihood model's profiles.

    Parameters
    ----------
    agg_obj : af.PhaseOutput
        A PyAutoFit aggregator's PhaseOutput object containing the generators of the results of model-fits.
    """

    masked_dataset = masked_dataset_from_agg_obj(
        agg_obj=agg_obj, settings_masked_dataset=settings_masked_dataset
    )

    xvalues = masked_dataset.xvalues
    instance = agg_obj.samples.max_log_likelihood_instance
    profiles = instance.profiles

    return sum([profile.profile_from_xvalues(xvalues=xvalues) for profile in profiles])


def fit_generator_from_aggregator(aggregator, settings_masked_dataset=None):
    """Compute a generator of *FitDataset* objects from an input aggregator, which generates a list of the
    *FitDataset* objects for every set of results loaded in the aggregator.

    This is performed by mapping the *fit_from_agg_obj* with the aggregator, which sets up each fit using
    only generators ensuring that manipulating the fits of large sets of results is done in a memory efficient way.

    Parameters
    ----------
    aggregator : af.Aggregator
        A PyAutoFit aggregator object containing the results of PyAutoGalaxy model-fits."""

    func = partial(fit_from_agg_obj, settings_masked_dataset=settings_masked_dataset)

    return aggregator.map(func=func)


def fit_from_agg_obj(agg_obj, settings_masked_dataset=None):
    """Compute a *Fit* object from an aggregator's *PhaseOutput* class, which we call an 'agg_obj' to describe  that it
    acts as the aggregator object for one result in the *Aggregator*. This uses the aggregator's generator outputs such
    that the function can use the *Aggregator*'s map function to to create a *Fit* generator.

     The *Fit* is created following the same method as the PyAutoGalaxy *Phase* classes. 

    Parameters
    ----------
    agg_obj : af.PhaseOutput
        A PyAutoFit aggregator's PhaseOutput object containing the generators of the results of PyAutoGalaxy model-fits.
    """
    masked_dataset = masked_dataset_from_agg_obj(
        agg_obj=agg_obj, settings_masked_dataset=settings_masked_dataset
    )
    model_data = model_data_from_agg_obj(
        agg_obj=agg_obj, settings_masked_dataset=settings_masked_dataset
    )

    return f.FitDataset(masked_dataset=masked_dataset, model_data=model_data)



from abc import ABC, abstractmethod
from typing import Dict, Callable, List

from .factor_graphs import Variable, Factor, FactorGraph, AbstractNode
from .mean_field import MeanFieldApproximation
from .messages import NormalMessage, AbstractMessage


class PriorVariable:
    def __init__(
            self,
            prior: Callable,
            variable: Variable
    ):
        """
        Comprises a prior and the variable to which it corresponds

        Parameters
        ----------
        prior
            A prior on the variable
        variable
            A placeholder for a variable
        """
        self.variable = variable
        self.factor = Factor(
            prior,
            x=variable
        )
        self.message = NormalMessage.from_prior(
            prior
        )


class AbstractMeanFieldPriorModel(ABC):
    def mean_field_approximation(
            self,
            variable_message_dict: Dict[Variable, AbstractMessage]
    ) -> MeanFieldApproximation:
        """
        Create a mean field approximation from this model.

        Parameters
        ----------
        variable_message_dict
            A dictionary mapping variables to messages. This is so messages corresponding
            to data can be passed in.

        Returns
        -------
        A mean-field approximation of this model
        """
        return MeanFieldApproximation.from_kws(
            self.model,
            {
                **self.message_dict,
                **variable_message_dict
            }
        )

    def __mul__(
            self,
            other: "AbstractMeanFieldPriorModel"
    ) -> "CompoundMeanFieldPriorModel":
        """
        Combine this model with another model

        Parameters
        ----------
        other
            Another model

        Returns
        -------
        A combined model
        """
        return CompoundMeanFieldPriorModel(
            self, other
        )

    @property
    @abstractmethod
    def model(self):
        pass

    @property
    @abstractmethod
    def message_dict(self):
        pass

    @property
    @abstractmethod
    def prior_variables(self):
        pass


class CompoundMeanFieldPriorModel(AbstractMeanFieldPriorModel):
    """
    A combination of two or more models
    """

    @property
    def model(self) -> FactorGraph:
        """
        Combine the factor graphs of the underlying models
        """
        model = self.mean_field_prior_models[0].model
        for mean_field_prior_model in self.mean_field_prior_models[1:]:
            model *= mean_field_prior_model.model
        return model

    @property
    def message_dict(self) -> Dict[Variable, AbstractMessage]:
        """
        Combine the message dictionaries of the underlying models
        """
        message_dict = dict()
        for mean_field_prior_model in self.mean_field_prior_models:
            message_dict.update(
                mean_field_prior_model.message_dict
            )
        return message_dict

    def __init__(self, *mean_field_prior_models):
        self.mean_field_prior_models = mean_field_prior_models

    @property
    def prior_variables(self) -> List[Variable]:
        """
        Combine the prior-variables of the underlying models
        """
        return [
            prior_variable
            for model
            in self.mean_field_prior_models
            for prior_variable
            in model.prior_variables
        ]


class MeanFieldPriorModel(AbstractMeanFieldPriorModel):
    def __init__(
            self,
            model: AbstractNode,
            **kwargs: Callable
    ):
        """
        Wraps a factor or graph

        Parameters
        ----------
        model
            A factor or graph
        kwargs
            Additional model components, such as priors
        """
        self._model = model
        self._variable_groups = dict()

        for name, item in kwargs.items():
            variable = getattr(
                model,
                name
            )
            group = PriorVariable(
                item,
                variable
            )
            self._variable_groups[
                name
            ] = group

    @property
    def model(self) -> FactorGraph:
        """
        Create a factor graph by combining the underlying model with any additional factors
        """
        model = self._model
        for group in self._variable_groups.values():
            model *= group.factor
        return model

    @property
    def message_dict(self) -> Dict[Variable, AbstractMessage]:
        """
        Create a message dict from all the underlying messages
        """
        return {
            group.variable: group.message
            for group
            in self._variable_groups.values()
        }

    @property
    def prior_variables(self) -> List[Variable]:
        """
        A list of variables contained in this model
        """
        return [
            group.variable
            for group
            in self._variable_groups.values()
        ]



import pytest

import autofit as af
from autofit import exc
from test_autofit import mock


@pytest.fixture(name="prior_1")
def make_prior_1():
    return af.UniformPrior()


@pytest.fixture(name="prior_2")
def make_prior_2():
    return af.UniformPrior()


@pytest.fixture(name="lower_assertion")
def make_lower_assertion(prior_1, prior_2):
    return prior_1 < prior_2


@pytest.fixture(name="greater_assertion")
def make_greater_assertion(prior_1, prior_2):
    return prior_1 > prior_2


def test_as_argument(
        prior_1,
        prior_2
):
    model = af.Collection(
        truth=prior_1 < prior_2
    )

    result = model.instance_for_arguments(
        {
            prior_1: 0,
            prior_2: 1
        }
    )
    assert result.truth is True

    result = model.instance_for_arguments(
        {
            prior_1: 1,
            prior_2: 0
        }
    )
    assert result.truth is False


class TestAssertion:
    def test_lower_equal_assertion(self, prior_1, prior_2):
        assertion = prior_1 <= prior_2
        assert assertion.instance_for_arguments({prior_1: 0.4, prior_2: 0.5}) is True
        assert assertion.instance_for_arguments({prior_1: 0.5, prior_2: 0.5}) is True
        assert assertion.instance_for_arguments({prior_1: 0.6, prior_2: 0.5}) is False

    def test_greater_equal_assertion(self, prior_1, prior_2):
        assertion = prior_1 >= prior_2
        assert assertion.instance_for_arguments({prior_1: 0.6, prior_2: 0.5}) is True
        assert assertion.instance_for_arguments({prior_1: 0.5, prior_2: 0.5}) is True

        assert assertion.instance_for_arguments({prior_1: 0.4, prior_2: 0.5}) is False

    def test_assert_on_arguments_lower(self, lower_assertion, prior_1, prior_2):
        assert lower_assertion.instance_for_arguments({prior_1: 0.3, prior_2: 0.5}) is True
        assert lower_assertion.instance_for_arguments({prior_1: 0.6, prior_2: 0.5}) is False

    def test_assert_on_arguments_greater(self, greater_assertion, prior_1, prior_2):
        assert greater_assertion.instance_for_arguments({prior_1: 0.6, prior_2: 0.5}) is True
        assert greater_assertion.instance_for_arguments({prior_1: 0.3, prior_2: 0.5}) is False

    def test_numerical_assertion(self, prior_1):
        assertion = prior_1 < 0.5

        assert assertion.instance_for_arguments({prior_1: 0.4}) is True
        assert assertion.instance_for_arguments({prior_1: 0.6}) is False

    def test_numerical_assertion_left(self, prior_1):
        assertion = 0.5 < prior_1

        assert assertion.instance_for_arguments({prior_1: 0.6}) is True
        assert assertion.instance_for_arguments({prior_1: 0.4}) is False
        assert assertion.instance_for_arguments({prior_1: 0.5}) is False

    def test_compound_assertion(self, prior_1):
        assertion = (0.2 < prior_1) < 0.5
        assert assertion.instance_for_arguments({prior_1: 0.3}) is True
        assert assertion.instance_for_arguments({prior_1: 0.1}) is False
        assert assertion.instance_for_arguments({prior_1: 0.6}) is False

    # noinspection PyUnresolvedReferences
    def test_annotation(self):
        model = af.PriorModel(mock.MockDistanceClass)
        assertion = model.one < model.two

        assert assertion.instance_for_arguments({
            model.one.value: 0.3,
            model.two.value: 0.4
        }) is True
        assert assertion.instance_for_arguments({
            model.one.value: 0.5,
            model.two.value: 0.4
        }) is False


@pytest.fixture(name="promise_model")
def make_promise_model(phase):
    return phase.result.model.one.component


@pytest.fixture(name="model")
def make_model(collection):
    return collection.last.model.one.component


class TestPromiseAssertion:
    def test_less_than(self, promise_model, collection, model):
        promise = promise_model.one < promise_model.two
        assert isinstance(promise, af.prior.GreaterThanLessThanAssertion)

        assertion = promise.populate(collection)
        assert isinstance(assertion, af.prior.GreaterThanLessThanAssertion)

    def test_greater_than(self, promise_model, collection, model):
        promise = promise_model.one > promise_model.two
        assert isinstance(promise, af.prior.GreaterThanLessThanAssertion)

    def test_greater_than_equal(self, promise_model, collection, model):
        promise = promise_model.one >= promise_model.two
        assert isinstance(promise, af.prior.GreaterThanLessThanEqualAssertion)

    def test_integer_promise_assertion(self, promise_model, collection, model):
        promise = promise_model.one > 1.0
        assert isinstance(promise, af.prior.GreaterThanLessThanAssertion)

    def test_compound_assertion(self, promise_model, collection, model):
        promise = (1.0 < promise_model.one) < 1.0
        assert isinstance(promise, af.prior.CompoundAssertion)

        assertion = promise.populate(collection)
        assert isinstance(assertion, af.prior.CompoundAssertion)


class TestModel:
    def test_assertion_in_model(self, prior_1, prior_2):
        model = af.ModelMapper()
        model.one = prior_1
        model.two = prior_2

        model.add_assertion(prior_1 < prior_2)

        model.instance_from_unit_vector([0.1, 0.2])
        with pytest.raises(af.exc.FitException):
            model.instance_from_unit_vector([0.2, 0.1])

    def test_populate_assertion_promises(self, promise_model, collection):
        model = af.ModelMapper()
        model.component = promise_model
        # noinspection PyTypeChecker
        model.add_assertion(promise_model.one < promise_model.two)

        model = model.populate(collection)

        assert model._assertions[0].instance_for_arguments(
            {
                model.component.one: 0.0,
                model.component.two: 1.0
            }
        ) is True
        assert model._assertions[0].instance_for_arguments(
            {
                model.component.one: 1.0,
                model.component.two: 0.0
            }
        ) is False

    def test_numerical(self):
        model = af.ModelMapper()
        model.add_assertion(True)
        model.instance_from_unit_vector([])

        model = af.ModelMapper()
        model.add_assertion(False)
        with pytest.raises(exc.FitException):
            model.instance_from_unit_vector([])



import numpy as np

import autofit as af
from autofit import expectation_propagation as mp
from .model import Gaussian, make_data, _gaussian, _likelihood

n_observations = 100

prior = af.GaussianPrior(
    mean=0,
    sigma=40
)


def test_gaussian():
    x = np.arange(n_observations)
    y = make_data(
        Gaussian(
            centre=50.0,
            intensity=25.0,
            sigma=10.0
        ),
        x
    )

    observations = mp.Plate(
        name="observations"
    )

    x_ = mp.Variable(
        "x", observations
    )
    y_ = mp.Variable(
        "y", observations
    )
    z = mp.Variable(
        "z", observations
    )

    gaussian = mp.Factor(
        _gaussian,
        x=x_
    ) == z

    likelihood = mp.Factor(
        _likelihood,
        z=z,
        y=y_
    )

    prior_model = mp.MeanFieldPriorModel(
        likelihood * gaussian,
        centre=prior,
        intensity=prior,
        sigma=prior
    )

    model_approx = prior_model.mean_field_approximation({
        x_: mp.FixedMessage(x),
        y_: mp.FixedMessage(y),
        z: mp.NormalMessage.from_mode(
            np.zeros(n_observations), 100
        ),
    })

    opt = mp.optimise.LaplaceOptimiser(
        model_approx,
        n_iter=3
    )

    opt.run()

    for variable in prior_model.prior_variables:
        print(f"{variable.name} = {opt.model_approx[variable].mu}")


def test_multiple():
    x = np.arange(n_observations)
    observations = mp.Plate(
        name="observations"
    )
    x_ = mp.Variable(
        "x", observations
    )
    intensity = mp.Variable(
        "intensity"
    )

    def make_messages_and_prior_model_for_gaussian(
            gaussian
    ):
        y = make_data(
            gaussian,
            x
        )

        y_ = mp.Variable(
            "y", observations
        )
        z = mp.Variable(
            "z", observations
        )

        gaussian = mp.Factor(
            _gaussian,
            x=x_,
            intensity=intensity
        ) == z

        likelihood = mp.Factor(
            _likelihood,
            z=z,
            y=y_
        )

        prior_model_ = mp.MeanFieldPriorModel(
            likelihood * gaussian,
            centre=prior,
            intensity=prior,
            sigma=prior
        )
        message_dict = {
            y_: mp.FixedMessage(y),
            z: mp.NormalMessage.from_mode(
                np.zeros(n_observations), 100
            )
        }

        return message_dict, prior_model_

    messages_0, prior_model_0 = make_messages_and_prior_model_for_gaussian(
        Gaussian(
            centre=40.0,
            intensity=25.0,
            sigma=20.0
        )
    )
    messages_1, prior_model_1 = make_messages_and_prior_model_for_gaussian(
        Gaussian(
            centre=50.0,
            intensity=25.0,
            sigma=10.0
        )
    )

    prior_model = prior_model_0 * prior_model_1

    assert len(set(prior_model.prior_variables)) == 5

    model_approx = prior_model.mean_field_approximation({
        x_: mp.FixedMessage(x),
        **messages_0,
        **messages_1
    })

    opt = mp.optimise.LaplaceOptimiser(
        model_approx,
        n_iter=3
    )

    opt.run()

    for variable in prior_model.prior_variables:
        print(f"{variable.name} = {opt.model_approx[variable].mu}")



import autofit as af

"""
The 'result.py' module is unchanged from the previous tutorial, although there is a short comment below worth 
reading.
"""


class Result(af.Result):
    def __init__(self, samples, previous_model, search, analysis):
        """
        The results of a non-linear search performed by a phase.

        Parameters
        ----------
        samples : af.Samples
            A class containing the samples of the non-linear search, including methods to get the maximum log
            likelihood model, errors, etc.
        analysis : Analysis
            The Analysis class used by this model-fit to fit the model to the data.
        """
        super().__init__(samples=samples, previous_model=previous_model, search=search)
        self.analysis = analysis

    @property
    def max_log_likelihood_model_data(self):

        """
        It is worth noting why we store the 'Analysis' class in the Result class. In this tutorial, we changed our
        model and how it created the model-data (e.g. as a sum of profiles). However, we did not need to change
        the result module in any way, because it uses the 'analysis.py' module.

        Had this function explicitly written out how the most likely model-data is created it would of needed to be
        updated, creating more work for ourselves!
        """

        return self.analysis.model_data_from_instance(
            instance=self.samples.max_log_likelihood_instance
        )

    @property
    def max_log_likelihood_fit(self):
        return self.analysis.fit_from_model_data(
            model_data=self.max_log_likelihood_model_data
        )



import autofit as af

from howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.dataset import (
    dataset as ds,
)

"""
This module create tags for phases settings that customize the analysis. We tag phases for two reasons:

    1) Tags describes the phase settings, making it explicit what analysis was used to create the results.

    2) Tags create unique output paths, ensuring that if you run multiple phases on the same data with different settings
       each non-linear search (e.g. Emcee) won't inadvertently use results generated via a different analysis method.
       
The settings of a _SettingsPhase_ use the settings objects of individual parts of the code. For example, below, it uses
as input a _SettingsMaskedDataset_ object, which in the module 'dataset/dataset.py' can be seen contains the inputs and
tags required for trimming data.

You may be surprised to see us using classes like _SettingsMaskedDataset_ and _SettingsPhase_, as the use of a class
may appear somewhat uncessary. We will discuss the use of classes and structure of the code at the end of the tutorial 
script.
"""


class SettingsPhase(af.AbstractSettingsPhase):
    def __init__(self, settings_masked_dataset=ds.SettingsMaskedDataset()):

        super().__init__()

        self.settings_masked_dataset = settings_masked_dataset

    """
    This function generates a string we'll use to 'tag' a phase which uses this setting, thus ensuring results are
    output to a unique path.
    
    It uses the tag of the _SettingsMaskedDataset_ to do this.
    """

    @property
    def tag(self):

        """You may well have many more tags which appear here."""

        return "settings" + self.settings_masked_dataset.tag



import os

import pytest

import autofit as af


class MedianPDFInstance:
    def __init__(self, name):
        self.name = name


class MockPhaseOutput:
    def __init__(self, directory, pipeline, phase, dataset):
        self.directory = directory
        self.pipeline = pipeline
        self.phase = phase
        self.dataset = dataset

    @property
    def median_pdf_instance(self):
        return MedianPDFInstance(
            self.phase
        )

    @property
    def output(self):
        return self


@pytest.fixture(name="aggregator")
def make_aggregator():
    aggregator = af.Aggregator("")
    aggregator.phases = [
        MockPhaseOutput(
            "directory/number/one",
            "pipeline1",
            "phase1",
            "dataset1"
        ),
        MockPhaseOutput(
            "directory/number/two",
            "pipeline1",
            "phase2",
            "dataset1"
        ),
        MockPhaseOutput(
            "directory/letter/a",
            "pipeline2",
            "phase2",
            "dataset2"
        ),
    ]
    return aggregator


@pytest.fixture(
    name="aggregator_directory"
)
def make_aggregator_directory():
    directory = os.path.dirname(
        os.path.realpath(__file__)
    )
    return f"{directory}/tools/files/aggregator"


@pytest.fixture(
    name="path_aggregator"
)
def make_path_aggregator(
        aggregator_directory
):
    aggregator = af.Aggregator(
        aggregator_directory
    )
    yield aggregator
    aggregator.remove_unzipped()


def test_completed_aggregator(
        aggregator_directory
):
    aggregator = af.Aggregator(
        aggregator_directory,
        completed_only=True
    )
    assert len(aggregator) == 1
    assert "completed" in aggregator[0].directory


class TestLoading:
    def test_unzip(self, path_aggregator):
        assert len(path_aggregator) == 2

    def test_pickles(self, path_aggregator):
        assert list(path_aggregator.values(
            "dataset"
        ))[0]["name"] == "dataset"
        assert list(path_aggregator.values(
            "model"
        ))[0]["name"] == "model"
        assert list(path_aggregator.values(
            "non_linear"
        ))[0]["name"] == "optimizer"
        assert list(path_aggregator.values(
            "nonsense"
        ))[0] is None


@pytest.fixture(
    name="ages_for_predicate"
)
def make_ages_for_predicate(path_aggregator):
    def ages_for_predicate(
            predicate
    ):
        result = path_aggregator.filter(
            predicate
        )
        return [child.age for child in result.values(
            "child"
        )]

    return ages_for_predicate


class TestNumericalFiltering:
    def test_equality(self, path_aggregator):
        predicate = path_aggregator.child.age == 17
        result = path_aggregator.filter(
            predicate
        )
        assert len(result) == 1
        assert list(result.values("child"))[0].age == 17

    def test_greater_than(
            self,
            ages_for_predicate,
            path_aggregator
    ):
        ages = ages_for_predicate(
            path_aggregator.child.age > 10
        )
        assert ages == [17]

    def test_greater_than_equal(
            self,
            ages_for_predicate,
            path_aggregator
    ):
        ages = ages_for_predicate(
            path_aggregator.child.age >= 10
        )
        assert set(ages) == {10, 17}

    def test_less_than(
            self,
            ages_for_predicate,
            path_aggregator
    ):
        ages = ages_for_predicate(
            path_aggregator.child.age < 11
        )
        assert ages == [10]

    def test_greater_than_rhs(
            self,
            ages_for_predicate,
            path_aggregator
    ):
        ages = ages_for_predicate(
            10 < path_aggregator.child.age
        )
        assert ages == [17]

    def test_less_than_rhs(
            self,
            ages_for_predicate,
            path_aggregator
    ):
        ages = ages_for_predicate(
            11 > path_aggregator.child.age
        )
        assert ages == [10]

    def test_aggregator_to_aggregator(self, path_aggregator):
        predicate = path_aggregator.child.age == path_aggregator.child.age
        assert len(path_aggregator.filter(predicate)) == 2

        predicate = path_aggregator.child.age > path_aggregator.child.age
        assert len(path_aggregator.filter(predicate)) == 0


class TestFiltering:
    def test_or(self, aggregator):
        predicate_one = aggregator.directory.contains(
            "one"
        )
        predicate_two = aggregator.directory.contains(
            "two"
        )
        result = aggregator.filter(
            predicate_one | predicate_two
        )
        assert len(result) == 2
        assert result.directories == [
            "directory/number/one",
            "directory/number/two"
        ]

    def test_and(self, aggregator):
        predicate_one = aggregator.pipeline == "pipeline1"
        predicate_two = aggregator.phase == "phase2"
        result = aggregator.filter(
            predicate_one & predicate_two
        )
        assert len(result) == 1
        assert result.directories == [
            "directory/number/two"
        ]

    def test_not_contains(self, aggregator):
        predicate = ~(aggregator.pipeline.contains("1"))
        result = aggregator.filter(
            predicate
        )
        assert len(result) == 1
        assert result[0].pipeline == "pipeline2"

    def test_not_equal(self, aggregator):
        predicate = aggregator.pipeline != "pipeline1"
        result = aggregator.filter(
            predicate
        )
        assert len(result) == 1
        assert result[0].pipeline == "pipeline2"

    def test_rhs(self, aggregator):
        predicate = "pipeline1" != aggregator.pipeline
        result = aggregator.filter(
            predicate
        )
        assert result.pipeline == ["pipeline2"]

        predicate = "pipeline1" == aggregator.pipeline
        result = aggregator.filter(
            predicate
        )
        assert result.pipeline == ["pipeline1", "pipeline1"]

    def test_filter_index(self, aggregator):
        assert list(aggregator.filter(
            aggregator.pipeline == "pipeline1"
        )[1:].values("pipeline")) == ["pipeline1"]
        assert list(aggregator[1:].filter(
            aggregator.pipeline == "pipeline1"
        ).values("pipeline")) == ["pipeline1"]

    def test_filter(self, aggregator):
        result = aggregator.filter(
            aggregator.pipeline == "pipeline1"
        )
        assert len(result) == 2
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline == "pipeline1",
            aggregator.phase == "phase1"
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline == "pipeline1"
        ).filter(
            aggregator.phase == "phase1"
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

    def test_filter_contains(self, aggregator):
        result = aggregator.filter(
            aggregator.pipeline.contains("1")
        )
        assert len(result) == 2
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline.contains("1"),
            aggregator.phase.contains("1")
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

        result = aggregator.filter(
            aggregator.pipeline.contains("1")
        ).filter(
            aggregator.phase == "phase1"
        )

        assert len(result) == 1
        assert result[0].pipeline == "pipeline1"

    def test_filter_contains_directory(self, aggregator):
        result = aggregator.filter(
            aggregator.directory.contains("number")
        )
        assert len(result) == 2

        result = aggregator.filter(
            aggregator.directory.contains("letter")
        )
        assert len(result) == 1


class TestOperations:

    def test_attribute(self, aggregator):
        assert list(
            aggregator.values("pipeline")
        ) == ["pipeline1", "pipeline1", "pipeline2"]
        assert list(
            aggregator.values("phase")
        ) == ["phase1", "phase2", "phase2"]
        assert list(
            aggregator.values("dataset")
        ) == ["dataset1", "dataset1", "dataset2"]

    def test_indexing(self, aggregator):
        assert list(
            aggregator[1:].values("pipeline")
        ) == ["pipeline1", "pipeline2"]
        assert list(
            aggregator[-1:].values("pipeline")
        ) == ["pipeline2"]
        assert aggregator[0].pipeline == "pipeline1"

    def test_group_by(self, aggregator):
        result = aggregator.group_by("pipeline")

        assert len(result) == 2
        assert len(result[0]) == 2
        assert len(result[1]) == 1

        result = result.filter(aggregator.phase == "phase2")

        assert len(result) == 2
        assert len(result[0]) == 1
        assert len(result[1]) == 1

        assert list(map(
            list,
            result.values("phase")
        )) == [["phase2"], ["phase2"]]

    def test_map(self, aggregator):
        def some_function(output):
            return f"{output.phase} {output.dataset}"

        results = aggregator.map(some_function)
        assert list(results) == [
            "phase1 dataset1",
            "phase2 dataset1",
            "phase2 dataset2",
        ]



# %%
"""
Tutorial 4: Source Code
=======================

Up to now, all tutorials have been self contained. That is, the code used to define the model, analysis, run the
non-linear search, load data and plot images were contained in the Jupyter notebooks or Python scritps you were run.

For a real software developmet project, this code woulnd't be contained in one script, but instead make up the
project's source code. In this tutorial, and every tutorial hereafter, we will set up our code as if it were
an actual software project with a clearly defined source code library. This source code can be found in the folder
'/autofit_workspace/howtofit/chapter_1_introduction/tutorial_4_source_code/src'.

Check it out, and first note the directory structure of the source code, which is separated into 5 packages: 'dataset',
'fit', 'model', 'plot' and 'phase'. This cleanly separates the different parts of the code which perform do different
tasks and I recommend your model-fitting project close adopts this structure!

For example, the code which handles the model is completely separate from the code which handles the analysis class.
The model thus never interfaces directly with PyAutoFit, ensuring good code design by removing dependencies between
parts of the code that do not need them! Its the same for the part of the code that stores data ('dataset')
and fits a model to a _Dataset_ ('fit') - by keeping them separate its clear which part of the code do what task.

This is a principle aspect of object oriented design and software engineering called 'separation of concerns' and all
templates we provide in the HowToFit series will adhere to it.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)


# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/howtofit/config",
    output_path=f"{workspace_path}/howtofit/output/chapter_1",  # <- This sets up where the non-linear search's outputs go.
)

# %%
"""
First, checkout the file 

 'autofit_workspace/howtofit/chapter_1_introduction/tutorial_4_source_code/src/__init__.py

Here, we have added imports to this file allowing us to import the entire project in one go, which we do below,
importing it as 'htf'. 

Many software projects tend not to do this, instead relying on the user explicitly importing every module in the 
project that need, for example:

 from tutorial_4_source_code.src.dataset.dataset import Dataset
 from tutorial_4_source_code.src.model.gaussian import Gaussian
 from tutorial_4_source_code.src.plot import dataset_plots
 from tutorial_4_source_code.src.plot import fit_plots

Clearly, this creates a burden on the user in them having to understand the project structure and a lot of unecessary
code that clutters their scripts! Furthermore, as you'll see below, by controlling the project import in this way
you can design an API that makes takes like plotting results more intuitive.
"""

# %%
from howtofit.chapter_1_introduction.tutorial_4_source_code import (
    src as htf,
)

# %%
"""
To begin, in the 'src' folder checkout the 'data' package, which contains one module, 'dataset.py'. Whereas before we 
had arrays which separately contained the data and noise_map, from here on we'll combine them into a 'Dataset' class, 
which can be easily extended if our model-fitting problem has additional data components.

To create the Dataset, we import the simulator module and use it to generate the Dataset's data and noise-map. 
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1

dataset = htf.Dataset(data=gaussian_x1.data, noise_map=gaussian_x1.noise_map)

# %%
"""
Previously, we manually specified how to plot the _Dataset_. These plotting functions are now in our source code, in the
'plot' package - check them out now! You'll note we have separate modules for plotting lines (e.g. anything which is 
line, the data, a residual-map, etc.), parts of the _Dataset_ or the results of a fit.

You should take note of two things:  

 - The _Dataset_ plot functions take instances of the _Dataset_ class, meaning we we don't have to manually the part of 
 our data we want to pass to the function, making for a more concise API.
 
 - In plot/__init__.py we have imported the 'dataset_plots', 'fit_plots' and 'line_plots' modules as their 
 corresponding class names; 'Dataset', 'FitDataset' and 'Line'. This again makes for a clean API, where it is 
 immediately obvious to the user how to plot the objects they have used elsewhere in the project for performing 
 calculations.

Lets use a plot function to plot our data.
"""

# %%
htf.plot.Dataset.data(dataset=dataset)
htf.plot.Dataset.noise_map(dataset=dataset)

# %%
"""
Next, look at the 'model' package, which contains the module 'gaussian.py'. This contains the _Gaussian_ class we have
used previous to compose the 1D _Gaussian_ model that we fit.

By packaging all the model components into a single package, this will make it straight forward to add new model
components to the source code.
"""

# %%
gaussian = htf.Gaussian(centre=50.0, intensity=2.0, sigma=20.0)

# %%
"""
Next, lets checkout the 'fit' package which contains the 'fit.py' module. This packages all the fitting methods we 
introduced in tutorial 2 into a single FitDataset class, making it straight forward to compute the residual-map, 
chi-squared map, log likelihood and so forth. 

Again, take note of how the fit plot functions take an instance of the FitDataset class and were imported as 
'FitDataset', making for a clean API where it is intuitive how to plot the fit.

Below, I used the _Gaussian_ model above to illustrate how we can easily plot different aspects of a fit. 
"""

# %%
model_data = gaussian.profile_from_xvalues(xvalues=dataset.xvalues)

fit = htf.FitDataset(dataset=dataset, model_data=model_data)

htf.plot.FitDataset.residual_map(fit=fit)
htf.plot.FitDataset.normalized_residual_map(fit=fit)
htf.plot.FitDataset.chi_squared_map(fit=fit)

# %%
"""
As we discussed in tutorial 2, the different steps of performing a fit (e.g. computing the residuals, the chi-squared,
log likelihood, and so forth) are pretty much generic tasks performed by any model-fitting problem. 

Thus, you should literally be able to copy and paste the FitDataset class found in this tutorial (and future tutorials) 
and use them in your modeling software! I have made sure the class works for datasets of higher dimensionality (e.g. 
2D images or 3D datacubes).
"""

# %%
"""
We're finally ready to look at how our source code sets up the non-linear search and model-fit. Whereas before, we 
manually set up the model, analysis and non-linear search, from now on we're going to use PyAutoFits 'phase API' which
uses the 'phase' package, which contains 3 modules: 'phase.py', 'analysis.py' and 'result.py'.

At this point, you should open and inspect (in detail) these 3 source code files. These are the heart of any PyAutoFit 
model fit! 

An over view of each is as follows:

phase.py -> contains the Phase class:

 - Receives the model to be fitted (in this case a single Gaussian).
 
 - Handles the directory structure of the output (in this example results are output to the folder
 '/output/phase_example/'.
 
  Is passed the data when run, which is set up for the analysis.

analysis.py -> contains the Analysis class (is a restructred version of the the previous tutorial's Analysis class):

 - Prepares the _Dataset_ for fitting.
 
 - Fits this _Dataset_ with a model instance to compute a log likelihood for every iteration of the non-linear search.

result.py -> contains the Result class:

 - Stores the Samples object containing information on the non-linear search's samples.
 
 - Has functions to create the model image, residual-map, chi-squared-map and so forth of the maximum log likelihood 
  model etc.
"""

# %%
"""
Perform a non-linear search in PyAutoFit now only requires that we instantiate and run a Phase object. The Phase 
performs the following tasks (which we performed manually in the previous tutorial):

 - Builds the model to be fitted and interfaces it with the non-linear search algorithm.
 
 - Receives the data to be fitted and prepares it so the model can fit it.
 
 - Contains the Analysis class that defines the log likelihood function.
 
 - Returns the results. including the non-linear search's samples and the maximum likelihood fit.

In the previous tutorial, after we composed our model using the _PriorModel_ object we had to manually specify its
priors. However, now we are using a source code, the priors are instead loaded from config files, specifically the
config file found at 'autofit_workspace/howtofit/config/json_piors/gaussian.json'. If you inspect this file, you'll 
see the priors are set up using the same values as the previous tutorial.

It is worth noting that the name of this config file, 'gaussian.json', is not a conincedence. It is named after the
module we imported to create the _PriorModel_, the 'gaussian.py' module. Thus, our the json_config files we use to
set up the default priors of different model components share the name of the module they are in! 

Although we don't in this tutorial, we could of course over write the priors with new priors as we did in the previous
tutorial.

Lets instantiate and run a phase, which reduces the task of performing a model-fit in PyAutoFit to just two lines. 
The results are output to the path 'autofit_workspace/howtofit/chapter_1_introduction/output/phase_t4/emcee', which in 
contrast to the previous tutorial includes the phase name in the path structure.
"""

# %%
phase = htf.Phase(
    phase_name="phase_t4", gaussian=af.PriorModel(htf.Gaussian), search=af.Emcee()
)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t4/emcee"
    " folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

result = phase.run(dataset=dataset)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
The phase returns a Result object, just like the model-fit performed in the previous tutorial did. However, in
'result.py' you'll have noted we extended the Results object to include a property containing an instance of the 
maximum likelihood fit.
"""

# %%
print(result.max_log_likelihood_fit)

# %%
"""
Another benefit of writing our plot functions so that their input is instances of class they plot is now clear. We can 
visualize our results by simply passing the instance which is readily available in the results to our plot functions!
"""

# %%
htf.plot.FitDataset.model_data(fit=result.max_log_likelihood_fit)
htf.plot.FitDataset.residual_map(fit=result.max_log_likelihood_fit)
htf.plot.FitDataset.chi_squared_map(fit=result.max_log_likelihood_fit)

# %%
"""
And with that, we have introduced the PyAutoFit phase API alongside an example project, which provides a template on
how to structure model-fitting software. 

All of the remaining tutorials will be provided with a 'src' source code folder, which we will add to the __init__.py
file of in order to design the API of our project. 

The functionality these tutorials describe will be reflected in the comments of the source code. At this point, you 
should be thinking about how you might wish to structure your model-fitting software!
"""



import autofit as af
from test_autofit import mock


def test_model_with_type(phase, collection):
    promise = phase.result.model[mock.MockComponents][0]

    result = promise.populate(collection)

    assert isinstance(result, af.PriorModel)
    assert result.cls == mock.MockComponents


def test_instance_with_type(phase, collection):
    promise = phase.result.instance[mock.MockComponents][0]

    result = promise.populate(collection)

    assert isinstance(result, mock.MockComponents)



import numpy as np

import autofit as af
from autofit import expectation_propagation as mp
from .model import Gaussian, make_data, _gaussian, _likelihood

n_observations = 100

prior = af.GaussianPrior(
    mean=0,
    sigma=40
)


def test_gaussian():
    x = np.arange(n_observations)
    y = make_data(
        Gaussian(
            centre=50.0,
            intensity=25.0,
            sigma=10.0
        ),
        x
    )

    observations = mp.Plate(
        name="observations"
    )

    # TODO: Can we derive variables by looking at function argument names?
    x_ = mp.Variable(
        "x", observations
    )
    y_ = mp.Variable(
        "y", observations
    )
    z = mp.Variable(
        "z", observations
    )
    centre = mp.Variable(
        "centre"
    )
    intensity = mp.Variable(
        "intensity"
    )
    sigma = mp.Variable(
        "sigma"
    )

    gaussian = mp.Factor(
        _gaussian,
        x=x_,
        centre=centre,
        intensity=intensity,
        sigma=sigma
    ) == z
    likelihood = mp.Factor(
        _likelihood,
        z=z,
        y=y_
    )

    # TODO: Can priors look like autofit priors? Could mp objects derive promise functionality from autofit?
    prior_centre = mp.Factor(
        prior,
        x=centre
    )
    prior_intensity = mp.Factor(
        prior,
        x=intensity
    )
    prior_sigma = mp.Factor(
        prior,
        x=sigma
    )

    model = likelihood * gaussian * prior_centre * prior_sigma * prior_intensity

    model_approx = mp.MeanFieldApproximation.from_kws(
        model,
        {
            centre: mp.NormalMessage.from_prior(
                prior
            ),
            intensity: mp.NormalMessage.from_prior(
                prior
            ),
            sigma: mp.NormalMessage.from_prior(
                prior
            ),
            x_: mp.FixedMessage(x),
            y_: mp.FixedMessage(y),
            z: mp.NormalMessage.from_mode(
                np.zeros(n_observations), 100
            ),
        }
    )

    opt = mp.optimise.LaplaceOptimiser(
        model_approx,
        n_iter=3
    )
    opt.run()

    for variable in centre, intensity, sigma:
        print(f"{variable.name} = {opt.model_approx[variable].mu}")



import numpy as np


class Gaussian:
    def __init__(
        self,
        centre=0.0,  # <- PyAutoFit recognises these constructor arguments
        intensity=0.1,  # <- are the Gaussian's model parameters.
        sigma=0.01,
    ):
        self.centre = centre
        self.intensity = intensity
        self.sigma = sigma

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the light profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        xvalues : ndarray
            The x coordinates in the original reference frame of the data.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )



import autofit as af
from howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.dataset.dataset import (
    Dataset,
    MaskedDataset,
)
from howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.phase.result import (
    Result,
)
from howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.phase.analysis import (
    Analysis,
)


"""The Phase class __init__ method is unchanged from the previous tutorial, however other methods are changed."""


class Phase(af.AbstractPhase):

    gaussian = af.PhaseProperty("gaussian")

    Result = Result

    @af.convert_paths
    def __init__(self, paths, gaussian, search):
        """
        A phase which fits a _Gaussian_ model using a non-linear search.

        Parameters
        ----------
        paths : af.Paths
            Handles the output directory structure.
        gaussian : gaussians.Gaussian
            The model component _Gaussian_ class fitted by this phase.
        search: class
            The class of a non_linear search
        """

        super().__init__(paths=paths, search=search)

        self.gaussian = gaussian

    """
    The run method is slightly different, as it now passed a mask in addition to the _Dataset_. These are used to set up
    the masked-dataset in the 'analysis.py' module.
    """

    def run(self, dataset: Dataset, mask):
        """
        Pass a _Dataset_ to the phase, running the phase and non-linear search.

        Parameters
        ----------
        dataset: aa.Dataset
            The _Dataset_ fitted by the phase, as defined in the 'dataset.py' module.
        mask: Mask
            The mask used for the analysis.

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising information on the non-linear search and the maximum likelihood model.
        """

        analysis = self.make_analysis(dataset=dataset, mask=mask)

        result = self.run_analysis(analysis=analysis)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset, mask):
        """
        Create an Analysis object, which creates the _Dataset_ and contains the functions which perform the fit.

        Parameters
        ----------
        dataset: aa.Dataset
            The _Dataset_ fitted by the phase, as defined in the 'dataset.py' module.

        Returns
        -------
        analysis : Analysis
            An analysis object that the non-linear search calls to determine the fit log_likelihood for a given model
            instance.
        """

        """To mask the _Dataset_ we simply pass both to the MaskedDataset class."""

        masked_dataset = MaskedDataset(dataset=dataset, mask=mask)

        """
        The 'image_path' is where visualizatiion of the model fit is output. Below, we direct it to the same path as
        the non-linear search output, but with an additional folder 'image' at the end. This path should be used
        for pretty much any project.
        """

        return Analysis(
            masked_dataset=masked_dataset, image_path=self.search.paths.image_path
        )

    def make_result(self, result, analysis):
        return self.Result(
            samples=result.samples,
            previous_model=self.model,
            search=self.search,
            analysis=analysis,
        )



import autofit as af
from howtofit.chapter_1_introduction.tutorial_4_source_code.src.fit import (
    fit as f,
)

"""
The 'analysis.py' module contains the Anasys class we introduced in tutorial 3. Whereas before this class took the
data and noise-map separately, we now pass it an instance of the _Dataset_ class. We have also restructured the code
performs the model-fit, so that we can reuse it in the 'result.py' module.
"""


class Analysis(af.Analysis):

    """In this tutorial the Analysis only contains the _Dataset_. More attributes will be included in later tutorials."""

    def __init__(self, dataset):

        super().__init__()

        self.dataset = dataset

    """
    In the log_likelihood_function below, 'instance' is an instance of our model, which in this tutorial we have
    defined as _Gaussian_ class in 'model.py'. This instance is a unit vector mapper via each parameters prior using
    unit values determined by the non-linear. Crucially, this gives us the instance of our model we need to fit our
    data!
    """

    """
    The reason a _Gaussian_ is mapped to an instance in this way is because of the following line in 'phase.py':

        gaussian = af.PhaseProperty("gaussian")

    Thus, PhaseProperties define our model components and thus tell the non-linear search what it is fitting! For
    your model-fitting problem, you'll need to update the PhaseProperty(s) to your model-components!
    """

    def log_likelihood_function(self, instance):
        """
        Determine the log likelihood of a fit of a _Gaussian_ to the dataset, using a model instance of the Gaussian.

        Parameters
        ----------
        instance : model.Gaussian
            The _Gaussian_ model instance.

        Returns
        -------
        fit : Fit.log_likelihood
            The log likelihood value indicating how well this model fit the _Dataset_.
        """
        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)
        return fit.log_likelihood

    def model_data_from_instance(self, instance):
        return instance.gaussian.profile_from_xvalues(xvalues=self.dataset.xvalues)

    def fit_from_model_data(self, model_data):
        return f.FitDataset(dataset=self.dataset, model_data=model_data)

    def visualize(self, instance, during_analysis):
        """
        Visualization will be covered in tutorial 4.

        Do not delete this, as PyAutoFit will crash otherwise!
        """

        pass



from inspect import getfullargspec
from itertools import chain, repeat
from typing import Tuple, Dict, Union, Set, NamedTuple, Callable

import numpy as np

from autofit.expectation_propagation.factor_graphs.abstract import AbstractNode, accept_variable_dict
from autofit.expectation_propagation.factor_graphs.numerical import numerical_jacobian
from autofit.expectation_propagation.factor_graphs.variable import Variable


class FactorValue(NamedTuple):
    """
    The value associated with some factor
    """

    log_value: np.ndarray
    deterministic_values: Dict[Variable, np.ndarray]


class Factor(AbstractNode):
    def __init__(
            self,
            factor: Callable,
            name=None,
            vectorised=False,
            **kwargs: Variable
    ):
        """
        A node in a graph representing a factor

        Parameters
        ----------
        factor
            A wrapper around some callable
        args
            Variables representing positional arguments for the function
        kwargs
            Variables representing keyword arguments for the function
        """
        self._name = name or factor.__name__
        self.vectorised = vectorised

        self._factor = factor
        self._deterministic_variables = set()

        args = getfullargspec(self._factor).args
        kwargs = {
            **kwargs,
            **{
                arg: Variable(arg)
                for arg
                in args
                if arg not in kwargs and arg != "self"
            }
        }

        self.__function_shape = None
        super().__init__(
            **kwargs
        )

    jacobian = numerical_jacobian

    @property
    def _variables(self):
        return set(self._kwargs.values())

    @property
    def name(self) -> str:
        return self._name

    @property
    def _kwargs_dims(self) -> Dict[str, int]:
        """
        The number of plates for each keyword argument variable
        """
        return {
            key: len(value)
            for key, value
            in self._kwargs.items()
        }

    @property
    def _variable_plates(self) -> Dict[str, np.ndarray]:
        """
        Maps the name of each variable to the indices of its plates
        within this node
        """
        return {
            variable: self._match_plates(
                variable.plates
            )
            for variable
            in self.all_variables
        }

    @property
    def n_deterministic(self) -> int:
        """
        How many deterministic variables are there associated with this node?
        """
        return len(self._deterministic_variables)

    def __hash__(self) -> int:
        return hash(self._factor)

    def _resolve_args(
            self,
            **kwargs: np.ndarray
    ) -> dict:
        """
        Transforms in the input arguments to match the arguments
        specified for the factor.

        Parameters
        ----------
        args
        kwargs

        Returns
        -------

        """
        return {n: kwargs[v.name] for n, v in self._kwargs.items()}

    def _function_shape(self, **kwargs) -> Tuple[int, ...]:
        """
        Calculates the expected function shape based on the variables
        """
        if self.__function_shape is None:
            kws = self._resolve_args(**kwargs)
            variables = dict(
                (self._kwargs[k], x) for k, x in kws.items())
            var_shapes = {v: np.shape(x) for v, x in variables.items()}
            var_dims_diffs = {
                v: len(s) - v.ndim
                for v, s in var_shapes.items()
            }
            """
            If all the passed variables have an extra dimension then 
            we assume we're evaluating multiple instances of the function at the 
            same time
    
            otherwise an error is raised
            """
            if set(var_dims_diffs.values()) == {1}:
                # Check if we're passing multiple values e.g. for sampling
                shift = 1
            elif set(var_dims_diffs.values()) == {0}:
                shift = 0
            else:
                raise ValueError("dimensions of passed inputs do not match")

            """
            Updating shape of output array to match input arrays
    
            singleton dimensions are always assumed to match as in
            standard array broadcasting
    
            e.g. (1, 2, 3) == (3, 2, 1)
            """
            shape = np.ones(self.ndim + shift, dtype=int)
            for v, vs in var_shapes.items():
                ind = self._variable_plates[v] + shift
                vshape = vs[shift:]
                if shift:
                    ind = np.r_[0, ind]
                    vshape = (vs[0],) + vshape

                if shape.size:
                    if not (
                            np.equal(shape[ind], 1) |
                            np.equal(shape[ind], vshape) |
                            np.equal(vshape, 1)).all():
                        raise AssertionError(
                            "Shapes do not match"
                        )
                    shape[ind] = np.maximum(shape[ind], vshape)

            self.__function_shape = tuple(shape)
        return self.__function_shape

    def _call_factor(
            self,
            **kwargs: Dict[str, np.ndarray]
    ) -> np.ndarray:
        """
        Call the underlying function

        Parameters
        ----------
        args
            Positional arguments for the function
        kwargs
            Keyword arguments for the function

        Returns
        -------
        Value returned by the factor
        """
        kws = self._resolve_args(
            **kwargs
        )

        if self.vectorised:
            return self._factor(**kws)
        return self._py_vec_call(**kws)

    def _py_vec_call(
            self,
            **kwargs: np.ndarray
    ) -> np.ndarray:
        """Some factors may not be vectorised to broadcast over
        multiple inputs

        this method checks whether multiple input values have been
        passed, and if so automatically loops over the inputs.
        If any of the inputs have initial dimension one, it repeats
        that value to match the length of the other inputs

        If the other inputs do not match then it raises ValueError
        """
        kwargs_dims = {k: np.ndim(a) for k, a in kwargs.items()}
        # Check dimensions of inputs directly match plates
        direct_call = (
            all(dim == kwargs_dims[k] for k, dim in self._kwargs_dims.items()))
        if direct_call:
            return self._factor(**kwargs)

        # Check dimensions of inputs match plates + 1
        vectorised = (
            all(dim + 1 == kwargs_dims[k]
                for k, dim in self._kwargs_dims.items()))

        if not vectorised:
            raise ValueError(
                "input dimensions do not match required dims"
                f"input: **kwargs={kwargs_dims}"
                f"required: "
                f"**kwargs={self._kwargs_dims}")

        kw_lens = {k: len(a) for k, a in kwargs.items()}

        # checking 1st dimensions match
        sizes = set(kw_lens.values())
        dim0 = max(sizes)
        if sizes.difference({1, dim0}):
            raise ValueError(
                f"size mismatch first dimensions passed: {sizes}")

        iter_kws = {
            k: iter(a) if kw_lens[k] == dim0 else iter(repeat(a[0]))
            for k, a in kwargs.items()}

        # iterator to generate keyword arguments
        def gen_kwargs():
            for i in range(dim0):
                yield {
                    k: next(a) for k, a in iter_kws.items()}

        # TODO this loop can also be parallelised for increased performance
        res = np.array([
            self._factor(**kws)
            for kws in gen_kwargs()])

        return res

    @accept_variable_dict
    def __call__(
            self,
            **kwargs: np.ndarray
    ) -> FactorValue:
        """
        Call the underlying factor

        Parameters
        ----------
        args
            Positional arguments for the factor
        kwargs
            Keyword arguments for the factor

        Returns
        -------
        Object encapsulating the result of the function call
        """
        val = self._call_factor(**kwargs)
        return FactorValue(
            val.reshape(
                self._function_shape(
                    **kwargs
                )
            ), {}
        )

    def broadcast_variable(
            self,
            variable: str,
            value: np.ndarray
    ) -> np.ndarray:
        """
        broadcasts the value of a variable to match the specific shape
        of the factor

        if the number of dimensions passed of the variable is 1
        greater than the dimensions of the variable then it's assumed
        that that dimension corresponds to multiple samples of that variable
        """
        return self._broadcast(
            self._variable_plates[variable],
            value
        )

    def collapse(
            self,
            variable: str,
            value: np.ndarray,
            agg_func=np.sum
    ) -> np.ndarray:
        """
        broadcasts the value of a variable to match the specific shape
        of the factor

        if the number of dimensions passed of the variable is 1
        greater than the dimensions of the variable then it's assumed
        that that dimension corresponds to multiple samples of that variable
        """
        ndim = np.ndim(value)
        shift = ndim - self.ndim
        assert shift in {0, 1}
        inds = self._variable_plates[variable] + shift
        dropaxes = tuple(np.setdiff1d(
            np.arange(shift, ndim), inds))

        # to ensured axes of returned array is in the correct order
        moved = np.moveaxis(value, inds, np.sort(inds))
        return agg_func(moved, axis=dropaxes)

    def __eq__(self, other: Union["Factor", Variable]):
        """
        If set equal to a variable that variable is taken to be deterministic and
        so a DeterministicFactorNode is generated.
        """
        if isinstance(other, Factor):
            if isinstance(other, type(self)):
                return (
                        (self._factor == other._factor)
                        and (frozenset(self._kwargs.items())
                             == frozenset(other._kwargs.items()))
                        and (frozenset(self.variables)
                             == frozenset(other.variables))
                        and (frozenset(self.deterministic_variables)
                             == frozenset(self.deterministic_variables)))
            else:
                return False

        from autofit.expectation_propagation.factor_graphs import DeterministicFactorNode
        return DeterministicFactorNode(
            self._factor,
            other,
            **self._kwargs
        )

    def __mul__(self, other):
        """
        When two factors are multiplied together this creates a graph
        """
        from autofit.expectation_propagation.factor_graphs.graph import FactorGraph
        return FactorGraph([self]) * other

    def __repr__(self) -> str:
        args = ", ".join(chain(
            map("{0[0]}={0[1]}".format, self._kwargs.items())))
        return f"Factor({self.name})({args})"

    @property
    def variables(self) -> Set[Variable]:
        """
        Dictionary mapping the names of variables to those variables
        """
        return self._variables

    @property
    def deterministic_variables(self) -> Set[Variable]:
        """
        Dictionary mapping the names of deterministic variables to those variables
        """
        return self._deterministic_variables



import numpy as np

from autofit import expectation_propagation as mp


def func(x):
    return 2 * x


def test_():
    x = mp.Variable("x")
    factor = mp.Factor(
        func,
        x=x
    )

    model = factor * factor

    value = model(x=np.array([1.0]))
    assert value.log_value == np.array([4.0])



import autofit as af
from howtofit.chapter_1_introduction.tutorial_4_source_code.src.dataset.dataset import (
    Dataset,
)
from howtofit.chapter_1_introduction.tutorial_4_source_code.src.phase.result import (
    Result,
)
from howtofit.chapter_1_introduction.tutorial_4_source_code.src.phase.analysis import (
    Analysis,
)

"""
The phase package combines a data-set, model and non-linear search, allowing us to fit the _Dataset_ with the model. It
essentially acts as the 'meeting point' between the other packages in the project (dataset, fit, plot) and modules
in the phase package (phase.py, analysis.py, result.py).
"""


class Phase(af.AbstractPhase):

    """
    This tells the phase that the input parameter 'gaussian' is a model component that is fitted for by the phase's
    non-linear search.

    In 'analysis.py', the function 'fit' has an input parameter called 'instance' which is the gaussian mapped from
    this model via a unit vector and the model priors (as described in tutorial 1).

    For your model-fitting problem, this will be replaced by the modules in your 'model' package.
    """

    gaussian = af.PhaseProperty("gaussian")

    Result = Result  # Set the result to the Result class in 'result.py'

    @af.convert_paths  # <- This handles setting up output paths.
    def __init__(
        self,
        paths,
        gaussian,  # <- The user inputs a model -> gaussian.py -> _Gaussian_ class here.
        search,  # <- This specifies the default non-linear search used by the phase.
    ):
        """
        A phase which fits a _Gaussian_ model using a non-linear search.

        Parameters
        ----------
        paths : af.Paths
            Handles the output directory structure.
        gaussian : model.gaussians.Gaussian
            The model component _Gaussian_ class fitted by this phase.
        search: class
            The class of a non_linear search
        """
        super().__init__(paths=paths, search=search)
        self.gaussian = gaussian

    def run(self, dataset: Dataset):
        """ Pass a _Dataset_ to the phase, running the phase and non-linear search.

        Parameters
        ----------
        _Dataset_ : _Dataset_.Dataset
            The _Dataset_ fitted by the phase, which is specified in the module 'dataset/dataset.py'

        Returns
        -------
        result: result.Result
            A result object comprising information on the non-linear search and the maximum likelihood model.
        """

        """
        These functions create instances of the Analysis class (in 'analysis.py'), runs the analysis (which performs
        the non-linear search ) and returns an instance of the Result class (in 'result.py').

        Once you've looked through this module, check those modules out to see exactly what these classes do!
        """

        analysis = self.make_analysis(dataset=dataset)

        """
        'run_analysis' is not located in analysis.py, instead it is an inherited method from the parent class
        'af.AbstractPhase'. Essentially, all this function does is begin the non-linear search, using the analysis
        created above.
        """

        result = self.run_analysis(analysis=analysis)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset):
        """
        Create an Analysis object, which uses the _Dataset_ with functions to perform a fit.

        Parameters
        ----------
        _Dataset_ : _Dataset_.Dataset
            The _Dataset_ fitted by the phase, which is specified in the module 'dataset/dataset.py'

        Returns
        -------
        analysis : Analysis
            An analysis object that the non-linear search calls to determine the fit log_likelihood for a given model
            instance.
        """
        return Analysis(dataset=dataset)

    def make_result(self, result, analysis):
        return self.Result(
            samples=result.samples,
            previous_model=self.model,
            search=self.search,
            analysis=analysis,
        )



import numpy as np
import pytest
from scipy import stats

from autofit import expectation_propagation as mp


@pytest.fixture(
    name="norm"
)
def make_norm():
    return stats.norm(loc=0, scale=1.)


@pytest.fixture(
    autouse=True
)
def set_seed():
    np.random.seed(1)


@pytest.fixture(
    name="prior_norm"
)
def make_prior_norm():
    return stats.norm(loc=0, scale=10.)


@pytest.fixture(
    name="prior"
)
def make_prior(prior_norm):
    def prior(x):
        return prior_norm.logpdf(x)

    return prior


def linear(x, a, b):
    return np.matmul(x, a) + np.expand_dims(b, -2)


@pytest.fixture(
    name="obs"
)
def make_obs():
    return mp.Plate(name='obs')


@pytest.fixture(
    name="features"
)
def make_features():
    return mp.Plate(name='features')


@pytest.fixture(
    name="dims"
)
def make_dims():
    return mp.Plate(name='dims')


@pytest.fixture(
    name="x_"
)
def make_x_(obs, features):
    return mp.Variable('x', obs, features)


@pytest.fixture(
    name="a_"
)
def make_a_(features, dims):
    return mp.Variable('a', features, dims)


@pytest.fixture(
    name="b_"
)
def make_b_(dims):
    return mp.Variable('b', dims)


@pytest.fixture(
    name="z_"
)
def make_z_(obs, dims):
    return mp.Variable('z', obs, dims)


@pytest.fixture(
    name="y_"
)
def make_y_(obs, dims):
    return mp.Variable('y', obs, dims)


@pytest.fixture(
    name="linear_factor"
)
def make_linear_factor(
        x_, a_, b_, z_
):
    return mp.Factor(linear, x=x_, a=a_, b=b_) == z_


@pytest.fixture(
    name="prior_a"
)
def make_prior_a(prior, a_):
    return mp.Factor(prior, x=a_)


@pytest.fixture(
    name="prior_b"
)
def make_prior_b(prior, b_):
    return mp.Factor(prior, x=b_)


@pytest.fixture(
    name="likelihood_factor"
)
def make_likelihood_factor(likelihood, z_, y_):
    return mp.Factor(likelihood, z=z_, y=y_)



from howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.plot import (
    dataset_plots,
    fit_plots,
)

"""
The visualizer is used by a phase to output results to the hard-disk. It is called both during the model-fit,
enabling the best-fit model of a non-linear search to be output on-the-fly (e.g. whilst it is still running) and
at the end.

The 'image_path' specifies the path where images are output. By default, this is the image_path of the search,
so the folder 'output/phase_name/image'.
"""


class AbstractVisualizer:
    def __init__(self, image_path):

        self.image_path = image_path


class Visualizer(AbstractVisualizer):
    def __init__(self, masked_dataset, image_path):

        """When the Visualizer is instantiated, the _MaskedDataset_ is passed to it and visualized."""

        super().__init__(image_path)

        self.masked_dataset = masked_dataset

        """For visualizing the dataset, we simmply plot the plot methods here."""

        dataset_plots.data(
            dataset=masked_dataset,
            output_filename="dataset_data",
            output_path=self.image_path,
            output_format="png",
        )
        dataset_plots.noise_map(
            dataset=masked_dataset,
            output_filename="dataset_noise_map",
            output_path=self.image_path,
            output_format="png",
        )

    def visualize_fit(self, fit, during_analysis):

        """
        The fit is visualized during the model-fit, thus it requires its own method which is called by the non-linear
        search every set number of intervals.

        Unlike the _Dataset_ plots above, the fit 'data' and 'noise-map' are masked.
        """

        fit_plots.data(
            fit=fit,
            output_filename="fit_data",
            output_path=self.image_path,
            output_format="png",
        )
        fit_plots.noise_map(
            fit=fit,
            output_filename="fit_noise_map",
            output_path=self.image_path,
            output_format="png",
        )
        fit_plots.model_data(
            fit=fit,
            output_filename="fit_model_data",
            output_path=self.image_path,
            output_format="png",
        )
        fit_plots.residual_map(
            fit=fit,
            output_filename="fit_residual_map",
            output_path=self.image_path,
            output_format="png",
        )
        fit_plots.chi_squared_map(
            fit=fit,
            output_filename="fit_chi_squared_map",
            output_path=self.image_path,
            output_format="png",
        )

        if not during_analysis:

            """
            If this function is called during an analysis, the during_analysis bool will be 'True'. If there are
            images you only want to output at the end of the analysis, you can thus save them for this if clause only.

            For example, this phase only visualizes individual images of the fit's normalized residual-map after the
            model fit has finished. For problems where there is a lot of visualization, and it can thus often take a
            long term to perform, limiting the output to after the analysis can be beneficial.
            """

            fit_plots.normalized_residual_map(
                fit=fit,
                output_filename="fit_normalized_residual_map",
                output_path=self.image_path,
                output_format="png",
            )



import numpy as np

"""The 'dataset.py' module is unchanged from the previous tutorial."""


class Dataset:
    def __init__(self, data, noise_map):
        """A class containing the data and noise-map of a 1D line _Dataset_.

        Parameters
        ----------
        data : np.ndarray
            The array of the data, in arbitrary units.
        noise_map : np.ndarray
            An array describing the RMS standard deviation error in each data pixel, in arbitrary units.
        """
        self.data = data
        self.noise_map = noise_map

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])


class MaskedDataset:
    def __init__(self, dataset, mask):
        """
        A masked dataset, which is an image, noise-map and mask.

        Parameters
        ----------
        dataset: im.Dataset
            The _Dataset_ (the image, noise-map, etc.)
        mask: msk.Mask
            The 1D mask that is applied to the _Dataset_.
        """

        """We store the unmasked _Dataset_ in the masked-dataset, incase we need it for anything."""
        self.dataset = dataset

        self.mask = mask

        """We apply the mask, setting all entries where the mask is True to zero."""
        self.data = dataset.data * np.invert(mask)

        """Same for the noise-map"""
        self.noise_map = dataset.noise_map * np.invert(mask)

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])

    def signal_to_noise_map(self):
        return self.data / self.noise_map



import inspect

from typing_inspect import is_tuple_type
from decorator import decorator


class DimensionType(float):
    def __new__(cls, value, *args, **kwargs):
        # noinspection PyArgumentList
        return float.__new__(cls, value)

    def __init__(self, value):
        float.__init__(value)


def _map_types(func, self, *args, **kwargs):
    annotations = inspect.getfullargspec(func).annotations

    def map_to_type(value, name=None, position=None):
        if isinstance(value, DimensionType):
            return value
        arg_type = None
        if name is not None:
            try:
                arg_type = annotations[name]
            except KeyError:
                pass
        if position is not None:
            arg_type = list(annotations.values())[position]
        if is_tuple_type(arg_type):
            return tuple(
                element_value
                if isinstance(element_value, DimensionType)
                else element_type(element_value)
                for element_type, element_value in zip(arg_type.__args__, value)
            )

        return arg_type(value)

    return func(
        self,
        *[map_to_type(value, position=index) for index, value in enumerate(args)],
        **{name: map_to_type(value, name=name) for name, value in kwargs.items()}
    )


map_types = decorator(_map_types)



import inspect
import autofit as af
# noinspection PyAbstractClass
from autofit.mapper.prior_model import attribute_pair


import math

class Circle:
    def __init__(self, radius):
        self.radius = radius

    def with_circumference(self, circumference):
        self.circumference = circumference

    @property
    def circumference(self):
        return self.radius * 2 * math.pi

    @circumference.setter
    def circumference(self, circumference):
        self.radius = circumference / (2 * math.pi)


class GeometryProfile:
    def __init__(self, centre=(0.0, 0.0)):
        """Abstract GeometryProfile, describing an object with y, x cartesian
        coordinates """
        self.centre = centre

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class SphericalProfile(GeometryProfile):
    def __init__(self, centre=(0.0, 0.0)):
        """ Generic circular profiles class to contain functions shared by light and
        mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profile.
        """
        super(SphericalProfile, self).__init__(centre)


class EllipticalProfile(SphericalProfile):
    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """ Generic elliptical profiles class to contain functions shared by light
        and mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalProfile, self).__init__(centre)
        self.axis_ratio = axis_ratio
        self.phi = phi


class MassProfile:
    def surface_density_func(self, eta):
        raise NotImplementedError("surface_density_at_radius should be overridden")

    def surface_density_from_grid(self, grid):
        pass

    def potential_from_grid(self, grid):
        pass

    def deflections_from_grid(self, grid):
        raise NotImplementedError("deflections_from_grid should be overridden")


# noinspection PyAbstractClass
class EllipticalMassProfile(EllipticalProfile, MassProfile):
    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """
        Abstract class for elliptical mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The origin of the profile
        axis_ratio : float
            Ellipse's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of profile's ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalMassProfile, self).__init__(centre=centre, axis_ratio=axis_ratio, phi=phi)
        self.axis_ratio = axis_ratio
        self.phi = phi


# noinspection PyAbstractClass
class EllipticalCoredIsothermal(EllipticalProfile):
    def __init__(
            self,
            centre=(0.0, 0.0),
            axis_ratio=1.0, phi=0.0,
            einstein_radius=1.0,
            core_radius=0.05,
    ):
        """
        Represents a cored elliptical isothermal density distribution, which is
        equivalent to the elliptical power-law
        density distribution for the value slope=2.0

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the origin of the profiles
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive
            x-axis
        einstein_radius : float
            Einstein radius of power-law mass profiles
        core_radius : float
            The radius of the inner core
        """

        super(EllipticalCoredIsothermal, self).__init__(
            centre=centre, axis_ratio=axis_ratio, phi=phi,
        )
        self.einstein_radius = einstein_radius
        self.core_radius = core_radius


class EllipticalSersic(EllipticalProfile):
    def __init__(
            self,
            centre=(0.0, 0.0),
            axis_ratio=1.0,
            phi=0.0,
            intensity=0.1,
            effective_radius=0.6,
            sersic_index=4.0,
    ):
        """ The elliptical Sersic profile, used for fitting a model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per
            second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concentration of the of the light profile.
        """
        super().__init__(
            centre=centre, axis_ratio=axis_ratio, phi=phi,
        )
        self.intensity = intensity
        self.effective_radius = effective_radius
        self.sersic_index = sersic_index


class EllipticalCoreSersic(EllipticalSersic):
    def __init__(
            self,
            centre=(0.0, 0.0),
            axis_ratio=1.0,
            phi=0.0,
            intensity=0.1,
            effective_radius=0.6,
            sersic_index=4.0,
            radius_break=0.01,
            intensity_break=0.05,
            gamma=0.25,
            alpha=3.0,
    ):
        """ The elliptical cored-Sersic profile, used for fitting a model_galaxy's
        light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per
            second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concetration of the of the light profile.
        radius_break : Float
            The break radius separating the inner power-law (with logarithmic slope
            gamma) and outer Sersic function.
        intensity_break : Float
            The intensity at the break radius.
        gamma : Float
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer
            Sersic profiles.
        """
        super(EllipticalCoreSersic, self).__init__(
            centre=centre,
            axis_ratio=axis_ratio,
            phi=phi,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index
        )
        self.radius_break = radius_break
        self.intensity_break = intensity_break
        self.alpha = alpha
        self.gamma = gamma


class EllipticalExponential(EllipticalSersic):
    def __init__(
            self,
            centre=(0.0, 0.0),
            axis_ratio=1.0,
            phi=0.0,
            intensity=0.1,
            effective_radius=0.6,
    ):
        """ The elliptical exponential profile, used for fitting a model_galaxy's light.

        This is a subset of the elliptical Sersic profile, specific to the case that
        sersic_index = 1.0.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per
            second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        """
        super(EllipticalExponential, self).__init__(
            centre=centre,
            axis_ratio=axis_ratio,
            phi=phi,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=1.0
        )



class EllipticalGaussian(EllipticalProfile):
    def __init__(
        self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, sigma=0.01
    ):
        """ The elliptical Gaussian profile.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per
            second).
        sigma : float
            The full-width half-maximum of the Gaussian.
        """
        super(EllipticalGaussian, self).__init__(centre, axis_ratio, phi)

        self.intensity = intensity
        self.sigma = sigma


class Redshift:
    def __init__(self, redshift):
        self.redshift = redshift


class Galaxy:
    def __init__(
            self,
            light_profiles: list = None,
            mass_profiles: list = None,
            redshift=None,
            **kwargs
    ):
        self.redshift = redshift
        self.light_profiles = light_profiles
        self.mass_profiles = mass_profiles
        self.kwargs = kwargs


class Tracer:
    def __init__(self, lens_galaxy: Galaxy, source_galaxy: Galaxy, grid):
        self.lens_galaxy = lens_galaxy
        self.source_galaxy = source_galaxy
        self.grid = grid

# noinspection PyAbstractClass
class GalaxyModel(af.AbstractPriorModel):
    def instance_for_arguments(self, arguments):
        try:
            return Galaxy(redshift=self.redshift.instance_for_arguments(arguments))
        except AttributeError:
            return Galaxy()

    def __init__(self, model_redshift=False, **kwargs):
        super().__init__()
        self.redshift = af.PriorModel(Redshift) if model_redshift else None
        print(self.redshift)
        self.__dict__.update(
            {
                key: af.PriorModel(value) if inspect.isclass(value) else value
                for key, value in kwargs.items()
            }
        )

    @property
    def instance_tuples(self):
        return []

    @property
    @attribute_pair.cast_collection(
        attribute_pair.PriorNameValue
    )
    def unique_prior_tuples(self):
        return (
            [item for item in self.__dict__.items() if isinstance(item[1], af.Prior)]
            + [("redshift", self.redshift.redshift)]
            if self.redshift is not None
            else []
        )

    @property
    @attribute_pair.cast_collection(af.PriorModelNameValue)
    def flat_prior_model_tuples(self):
        return [
            item
            for item in self.__dict__.items()
            if isinstance(item[1], af.AbstractPriorModel)
        ]


import autofit as af

from howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.fit import (
    fit as f,
)
from howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.phase import (
    visualizer,
)

"""The 'analysis.py' module in this tutorial is unchanged from the previous tutorial."""


class Analysis(af.Analysis):
    def __init__(self, masked_dataset, image_path=None):

        super().__init__()

        """The _MaskedDataset_ and visualizer are created in the same way as tutorial 6."""

        self.masked_dataset = masked_dataset

        self.visualizer = visualizer.Visualizer(
            masked_dataset=self.masked_dataset, image_path=image_path
        )

    def log_likelihood_function(self, instance):
        """Determine the fit of a list of Profiles (Gaussians, Exponentials, etc.) to the dataset, using a
        model instance.

        Parameters
        ----------
        instance
            The list of Profile model instance (e.g. the Gaussians, Exponentials, etc.).

        Returns
        -------
        fit : Fit.log_likelihood
            The log likelihood value indicating how well this model fit the _MaskedDataset_.
        """
        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)
        return fit.log_likelihood

    def model_data_from_instance(self, instance):
        return sum(
            [
                profile.profile_from_xvalues(xvalues=self.masked_dataset.xvalues)
                for profile in instance.profiles
            ]
        )

    def fit_from_model_data(self, model_data):
        return f.FitDataset(masked_dataset=self.masked_dataset, model_data=model_data)

    def visualize(self, instance, during_analysis):

        model_data = self.model_data_from_instance(instance=instance)
        fit = self.fit_from_model_data(model_data=model_data)

        self.visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)



from .dataset.dataset import Dataset, SettingsMaskedDataset, MaskedDataset
from .model import profiles
from .fit.fit import FitDataset
from . import plot
from .phase.phase import Phase
from .phase.settings import SettingsPhase



from os import path

import autofit as af
from autoconf import conf
from autofit.non_linear.mock.mock_search import MockSearch, MockAnalysis, MockSamples
from test_autofit import mock

directory = path.dirname(path.realpath(__file__))

conf.instance = conf.Config(
    config_path=path.join(directory, "files/config"),
)


def make_pipeline_1(name, folders, search):
    phase = af.Phase(
        phase_name="phase_1",
        folders=folders,
        model=af.PriorModel(
            mock.MockComponents,
            parameter=af.GaussianPrior(10.0, 1.0)
        ),
        search=search,
        analysis_class=MockAnalysis,
    )
    return af.Pipeline(f"{name}_1", phase)


def make_pipeline_2(name, folders, search):
    phase = af.Phase(
        phase_name="phase_2",
        folders=folders,
        model=af.PriorModel(
            mock.MockComponents,
            parameter=af.last.model.parameter
        ),
        search=search,
        analysis_class=MockAnalysis,
    )
    return af.Pipeline(f"{name}_2", phase)


def make_pipeline(
        name,
        folders=tuple(),
        search=MockSearch(
            samples=MockSamples(
                gaussian_tuples=[(0.5, 0.5)]
            )
        )
):
    pipeline_2 = make_pipeline_2(
        name,
        folders,
        search
    )
    pipeline_1 = make_pipeline_1(
        name,
        folders,
        search
    )

    return pipeline_1 + pipeline_2


def test_pipeline_composition():
    pipeline = make_pipeline("test")
    results = pipeline.run(
        mock.MockDataset()
    )
    assert results[0].model.parameter.mean == results[1].model.parameter.mean



from typing import (
    Dict, Tuple, Iterator
)

import numpy as np

from .abstract import AbstractMessage
from .fixed import FixedMessage
from .gamma import GammaMessage
from .normal import NormalMessage


def map_dists(dists: Dict[str, AbstractMessage],
              values: Dict[str, np.ndarray],
              _call: str = 'logpdf'
              ) -> Iterator[Tuple[str, np.ndarray]]:
    """
    Calls a method (default: logpdf) for each Message in dists
    on the corresponding value in values
    """
    for v in dists.keys() & values.keys():
        dist = dists[v]
        if isinstance(dist, AbstractMessage):
            yield v, getattr(dist, _call)(values[v])



from configparser import NoOptionError
import logging

from autoconf import conf

logger = logging.getLogger(__name__)


def parameter_labels_from_model(model) -> [str]:
    """A list of every parameter's label, used by *GetDist* for model estimation and visualization.

    The parameter labels are determined using the label.ini and label_format.ini config files."""

    parameter_labels = []
    prior_class_dict = model.prior_class_dict

    for prior_name, prior in model.prior_tuples_ordered_by_id:
        try:
            param_string = conf.instance.label.label(prior_name)
        except NoOptionError:
            logger.warning(
                f"No label provided for {prior_name}. Using prior name instead."
            )
            param_string = prior_name

        cls = prior_class_dict[prior]
        cls_string = conf.instance.label.subscript(cls)

        parameter_label = "{}_{{\\mathrm{{{}}}}}".format(param_string, cls_string)
        parameter_labels.append(parameter_label)

    return parameter_labels


from collections import Counter, defaultdict
from typing import Tuple, Dict, Collection, List, Callable

import numpy as np

from autofit.expectation_propagation.factor_graphs import FactorValue, AbstractNode
from autofit.expectation_propagation.factor_graphs.abstract import accept_variable_dict
from autofit.expectation_propagation.factor_graphs.factor import Factor
from autofit.expectation_propagation.factor_graphs.variable import Variable, Plate
from autofit.expectation_propagation.utils import add_arrays


class DeterministicFactorNode(Factor):
    def __init__(
            self,
            factor: Callable,
            variable: Variable,
            *args: Variable,
            **kwargs: Variable
    ):
        """
        A deterministic factor is used to convert a function f(g(x)) to f(y)g(x) (integrating over y wit
        a delta function) so that it can be represented in a factor graph.

        Parameters
        ----------
        factor
            The original factor to which the deterministic factor is associated
        variable
            The deterministic factor used
        args
            Variables for the original factor
        kwargs
            Variables for the original factor
        """
        super().__init__(
            factor,
            *args,
            **kwargs
        )
        self._deterministic_variables = {
            variable
        }

    @accept_variable_dict
    def __call__(
            self,
            **kwargs: np.ndarray
    ) -> FactorValue:
        """
        Call this factor with a set of arguments

        Parameters
        ----------
        args
            Positional arguments for the underlying factor
        kwargs
            Keyword arguments for the underlying factor

        Returns
        -------
        An object encapsulating the value for the factor
        """
        res = self._call_factor(**kwargs)
        shape = self._function_shape(**kwargs)
        shift = len(shape) - self.ndim
        plate_dim = dict(zip(self.plates, shape[shift:]))

        det_shapes = {
            v: shape[:shift] + tuple(
                plate_dim[p] for v in self.deterministic_variables
                for p in v.plates)
            for v in self.deterministic_variables
        }

        if not (isinstance(res, tuple) or self.n_deterministic > 1):
            res = res,

        log_val = 0. if shape == () else np.zeros(np.ones_like(shape))
        det_vals = {
            k: np.reshape(val, det_shapes[k])
            if det_shapes[k]
            else val
            for k, val
            in zip(self._deterministic_variables, res)
        }
        return FactorValue(log_val, det_vals)

    def __repr__(self) -> str:
        factor_str = super().__repr__()
        var_str = ", ".join(sorted(variable.name for variable in self._deterministic_variables))
        return f"({factor_str} == ({var_str}))"


class FactorGraph(AbstractNode):
    def __init__(
            self,
            factors: Collection[Factor],
    ):
        """
        A graph relating factors

        Parameters
        ----------
        factors
            Nodes wrapping individual factors in a model
        """
        self._name = ".".join(f.name for f in factors)

        self._factors = tuple(factors)

        self._factor_all_variables = {
            f: f.all_variables for f in self._factors
        }

        self._call_sequence = self._get_call_sequence()

        self._validate()

        _kwargs = {
            variable.name: variable
            for variable
            in self.variables
        }

        super().__init__(
            **_kwargs
        )

    @property
    def deterministic_variables(self):
        return self._deterministic_variables

    def broadcast_plates(
            self,
            plates: Collection[Plate],
            value: np.ndarray
    ) -> np.ndarray:
        """
        Extract the indices of a collection of plates then match
        the shape of the data to that shape.

        Parameters
        ----------
        plates
            Plates representing the dimensions of some factor
        value
            A value to broadcast

        Returns
        -------
        The value reshaped to match the plates
        """
        return self._broadcast(self._match_plates(plates), value)

    @property
    def name(self):
        return self._name

    def _validate(self):
        """
        Raises
        ------
        If there is an inconsistency with this graph
        """
        det_var_counts = ", ".join(
            v for v, c in Counter(
                v for f in self.factors
                for v in f.deterministic_variables).items()
            if c > 1)
        if det_var_counts:
            raise ValueError(
                "Improper FactorGraph, "
                f"Deterministic variables {det_var_counts} appear in "
                "multiple factors"
            )

    @property
    def _variables(self):
        return {
            variable
            for factor
            in self.factors
            for variable
            in factor.variables
        }

    @property
    def _deterministic_variables(self):
        return {
            variable
            for factor
            in self.factors
            for variable
            in factor.deterministic_variables
        }

    @property
    def variables(self):
        return self._variables - self._deterministic_variables

    def _get_call_sequence(self) -> List[List[Factor]]:
        """
        Compute the order in which the factors must be evaluated. This is done by checking whether
        all variables required to call a factor are present in the set of variables encapsulated
        by all factors, not including deterministic variables.

        Deterministic variables must be computed before the dependent factors can be computed.
        """
        call_sets = defaultdict(list)
        for factor in self.factors:
            missing_vars = frozenset(factor.variables.difference(self.variables))
            call_sets[missing_vars].append(factor)

        call_sequence = []
        while call_sets:
            # the factors that can be evaluated have no missing variables
            factors = call_sets.pop(frozenset(()))
            # if there's a KeyError then the FactorGraph is improper
            calls = []
            new_variables = set()
            for factor in factors:
                if isinstance(factor, DeterministicFactorNode):
                    det_vars = factor.deterministic_variables
                else:
                    det_vars = set()

                calls.append(factor)
                new_variables.update(det_vars)

            call_sequence.append(calls)

            # update to include newly calculated factors
            for missing in list(call_sets.keys()):
                if missing.intersection(new_variables):
                    factors = call_sets.pop(missing)
                    call_sets[missing.difference(new_variables)].extend(factors)

        return call_sequence

    @accept_variable_dict
    def __call__(
            self,
            **kwargs: np.ndarray
    ) -> FactorValue:
        """
        Call each function in the graph in the correct order, adding the logarithmic results.

        Deterministic values computed in initial factor calls are added to a dictionary and
        passed to subsequent factor calls.

        Parameters
        ----------
        args
            Positional arguments
        kwargs
            Keyword arguments

        Returns
        -------
        Object comprising the log value of the computation and a dictionary containing
        the values of deterministic variables.
        """

        # generate set of factors to call, these are indexed by the
        # missing deterministic variables that need to be calculated
        log_value = 0.
        det_values = {}
        variables = kwargs

        missing = set(self.kwarg_names) - variables.keys()
        if missing:
            n_miss = len(missing)
            missing_str = ", ".join(missing)
            raise ValueError(
                f"{self} missing {n_miss} arguments: {missing_str}"
                f"factor graph call signature: {self.call_signature}"
            )

        for calls in self._call_sequence:
            # TODO parallelise this part?
            for factor in calls:
                ret = factor(variables)
                ret_value = self.broadcast_plates(factor.plates, ret.log_value)
                log_value = add_arrays(log_value, ret_value)
                det_values.update(ret.deterministic_values)
                variables.update(ret.deterministic_values)

        return FactorValue(log_value, det_values)

    def __mul__(self, other: AbstractNode) -> "FactorGraph":
        """
        Combine this object with another factor node or graph, creating
        a new graph that comprises all of the factors of the two objects.
        """
        factors = self.factors

        if isinstance(other, FactorGraph):
            factors += other.factors
        elif isinstance(other, Factor):
            factors += (other,)
        else:
            raise TypeError(
                f"type of passed element {(type(other))} "
                "does not match required types, (`FactorGraph`, `FactorNode`)")

        return type(self)(factors)

    def __repr__(self) -> str:
        factors_str = " * ".join(map(repr, self.factors))
        return f"({factors_str})"

    @property
    def factors(self) -> Tuple[Factor, ...]:
        return self._factors

    @property
    def factor_all_variables(self) -> Dict[Factor, List[Variable]]:
        return self._factor_all_variables



from howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.plot import (
    line_plots,
)

"""
To visualize images during a phase, we need to be able to output them to hard-disk as a file (e.g a .png'). The line
plot function and fit plot functions below have been extended to provide this functionality.

The benefit of writing our visualization methods in this style, as separate functions in a specific 'plot' module is
now more clear. In 'visualizer.py', this makes it a lot more straight forward to plot each component of the fit.
However, the real benefit of this style will become fully apparently in tutorial 6.
"""


def data(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the data values of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit _Dataset_ whose data is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.data,
        ylabel="Data Values",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def noise_map(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the noise-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose noise-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.noise_map,
        ylabel="Noise Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def signal_to_noise_map(
    fit, output_path=None, output_filename=None, output_format="show"
):
    """Plot the signal-to-noise-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose signal-to-noise-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.signal_to_noise_map,
        ylabel="Signal-To-Noise Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def model_data(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the model data of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The Fit model data.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.model_data,
        ylabel="Model Data",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def residual_map(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the residual-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose residual-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.residual_map,
        ylabel="Residual Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def normalized_residual_map(
    fit, output_path=None, output_filename=None, output_format="show"
):
    """Plot the normalized_residual-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose normalized residual-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.normalized_residual_map,
        ylabel="Normalized Residual Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def chi_squared_map(fit, output_path=None, output_filename=None, output_format="show"):
    """Plot the chi_squared-map of a Fit.

    Parameters
    -----------
    Fit : fit.Fit
        The observed Fit whose chi-squared-map is plotted.
    """
    line_plots.line(
        xvalues=fit.xvalues,
        line=fit.chi_squared_map,
        ylabel="Chi-Squared Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
